import datetime
import traceback
import decimal
from six.moves.collections_abc import Iterable

from types import TracebackType

import six

from google.protobuf.internal.type_checkers import Int64ValueChecker
from rook.processor.namespace_serializer_base import NamespaceSerializerBase

from .namespaces.container_namespace import ContainerNamespace
from .namespaces.python_object_namespace import PythonObjectNamespace
from .namespaces.collection_namespace import ListNamespace, LIST_TYPE
from .namespaces.traceback_namespace import TracebackNamespace
from .namespaces.error_namespace import ErrorNamespace
from .namespaces.formatted_namespace import FormattedNamespace

from rook.logger import logger

from rook.protobuf import variant_pb2, variant2_pb2

from ..user_warnings import UserWarnings


class NamespaceSerializer2(NamespaceSerializerBase):
    def __init__(self):
        NamespaceSerializerBase.__init__(self, True)
        self.buffer_cache = {}

    def dump(self, namespace, variant, log_errors=True):
        try:
            if isinstance(namespace, ContainerNamespace):
                self._dump_container_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, PythonObjectNamespace):
                self._dump_object_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, ErrorNamespace):
                self._dump_error_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, FormattedNamespace):
                self._dump_formatted_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, TracebackNamespace):
                self._dump_traceback_namespace(namespace, variant, log_errors)
            else:
                raise NotImplementedError("Does not support serializing this type!", type(namespace))
        except Exception as e:
            message = "Failed to serialize namespace"

            variant.Clear()
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_ERROR)

            if log_errors:
                from .error import Error
                logger.exception(message)
                UserWarnings.send_warning(Error(exc=e, message=message))

    def dumps(self, namespace, log_errors=True):
        variant = variant2_pb2.Variant2()
        self.dump(namespace, variant, log_errors)
        return variant

    def _dump_container_namespace(self, namespace, variant, log_errors):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_NAMESPACE)

        for key, value in six.iteritems(namespace.dictionary):
            variant.attribute_names_in_cache.append(self._get_string_index_in_cache(key))
            attribute_value = variant.attribute_values.add()
            self.dump(value, attribute_value, log_errors)

            self.estimated_pending_bytes += 4  # One number (packed field), One header + length

    def _dump_object_namespace(self, namespace, variant, log_errors):
        self._dump_python_object(namespace.obj, variant, 0, namespace.dump_config, log_errors)

    def _dump_formatted_namespace(self, namespace, variant, log_errors):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_FORMATTED_MESSAGE)

        variant.bytes_index_in_cache = self._get_string_index_in_cache(namespace.obj)
        self.estimated_pending_bytes += 3  # Header + number

    def _dump_python_object(self, obj, variant, current_depth, config, log_errors):
        try:
            self._dump_python_object_unsafe(obj, variant, current_depth, config, log_errors)
        except Exception as e:
            message = "Failed to serialize namespace"

            variant.Clear()
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_ERROR)

            if log_errors:
                from .error import Error
                logger.exception(message)
                UserWarnings.send_warning(Error(exc=e, message=message))

    def _dump_python_object_unsafe(self, obj, variant, current_depth, config, log_errors):
        original_type = type(obj).__name__
        variant.original_type_index_in_cache = self._get_string_index_in_cache(original_type)
        self.estimated_pending_bytes += 3  # Header + number

        if isinstance(obj, NamespaceSerializerBase.PRIMITIVE_TYPES):
            self._dump_primitive(obj, original_type, variant, config.max_string)
        elif isinstance(obj, LIST_TYPE):
            self._dump_list(obj, variant, current_depth, config, log_errors)
        elif isinstance(obj, dict):
            self._dump_dictionary(obj, variant, current_depth, config, log_errors)
        elif isinstance(obj, BaseException):
            self._dump_exception(obj, variant, current_depth, config, log_errors)
        elif isinstance(obj, TracebackType):
            self._dump_traceback(obj, variant, current_depth, config, log_errors)
        elif NamespaceSerializer2.is_numpy_obj(obj):
            self._dump_primitive(obj.item(), original_type, variant, config.max_string)
        elif NamespaceSerializer2.is_torch_obj(obj):
            self._dump_primitive(str(obj), original_type, variant, config.max_string)
        elif NamespaceSerializer2.is_multidict_obj(obj):
            self._dump_primitive(str(obj), original_type, variant, config.max_string)
        elif NamespaceSerializer2.is_protobuf_obj(obj):
            self._dump_protobuf(obj, variant, current_depth, config, log_errors)
        elif hasattr(obj, '__dict__'):
            self._dump_user_class(obj, variant, current_depth, config, log_errors)
        else:
            self._dump_not_supported(obj, variant)

    def _dump_traceback(self, obj, variant, current_depth, config, log_errors):
        # python separates the "forward" stack (callees of the except clause)
        # and the "backward" stack (callers of above)
        # Possibly would be more useful to wrap this in a StackNamespace
        tb = traceback.format_tb(obj)
        tb[1:1] = traceback.format_stack(obj.tb_frame.f_back)
        value = ''.join(tb)

        self.dump_variant_type(variant, variant.VARIANT_STRING)

        variant.original_size = len(value)
        variant.bytes_index_in_cache = self._get_string_index_in_cache(value)
        self.estimated_pending_bytes += 6  # Header + number + header + number

    def _dump_primitive(self, obj, original_type, variant, max_string):
        if obj is None:
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_NONE)

        elif isinstance(obj, six.integer_types) and (obj < Int64ValueChecker._MAX) and (obj > Int64ValueChecker._MIN):
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_LONG)
            variant.long_value = int(obj)
            self.estimated_pending_bytes += 3  # Header + number

        elif isinstance(obj, six.integer_types):
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_LARGE_INT)
            variant.bytes_index_in_cache = self._get_string_index_in_cache(str(obj))
            self.estimated_pending_bytes += 3  # Header + number

        elif isinstance(obj, bool):
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_LONG)
            variant.long_value = int(obj)
            self.estimated_pending_bytes += 2  # Header + short number

        elif isinstance(obj, float):
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_DOUBLE)
            variant.double_value = float(obj)
            self.estimated_pending_bytes += 7  # Header + 64 bit float

        elif isinstance(obj, decimal.Decimal):
            serialized_decimal = str(obj)

            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_STRING)
            variant.original_size = len(serialized_decimal)
            variant.bytes_index_in_cache = self._get_string_index_in_cache(str(serialized_decimal))
            self.estimated_pending_bytes += 6  # Header + number + header + number

        elif isinstance(obj, six.string_types):
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_STRING)
            variant.original_size = len(obj)

            if len(obj) > max_string:
                obj = obj[:max_string]

            string = self.normalize_string(obj)
            variant.bytes_index_in_cache = self._get_string_index_in_cache(string)
            self.estimated_pending_bytes += 6  # Header + number + header + number

        elif isinstance(obj, self.BINARY_TYPES) or original_type == 'binary_type':
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_BINARY)
            variant.original_size = len(obj)

            if len(obj) > max_string:
                obj = obj[:max_string]

            variant.bytes_index_in_cache = self._get_bytes_index_in_cache(bytes(obj))
            self.estimated_pending_bytes += 6  # Header + number + header + number

        elif isinstance(obj, self.CODE_TYPES):
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_CODE_OBJECT)
            code_value = variant.code_values.add()
            code_value.name_index_in_cache = self._get_string_index_in_cache(self.normalize_string(obj.__name__))
            if hasattr(obj, '__code__') and hasattr(obj.__code__, 'co_filename'):
                code_value.filename_index_in_cache = self._get_string_index_in_cache(self.normalize_string(obj.__code__.co_filename))
                code_value.lineno = int(obj.__code__.co_firstlineno)
            if hasattr(obj, '__module__') and obj.__module__:
                code_value.module_index_in_cache = self._get_string_index_in_cache(self.normalize_string(obj.__module__))

            self.estimated_pending_bytes += 14  # Header + size + (header + number) * 4

        elif isinstance(obj, complex):
            self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_COMPLEX)
            variant.complex_value.real = float(obj.real)
            variant.complex_value.imaginary = float(obj.imag)
            self.estimated_pending_bytes += 8  # Large header + size + (header + 64 bit float) * 2

        elif isinstance(obj, datetime.datetime):
            self._dump_datetime(obj, variant)

        else:
            raise ValueError("Object is not a supported primitive!", type(obj))

    def _dump_datetime(self, obj, variant):
        if obj.tzinfo:
            obj = obj.replace(tzinfo=None)

        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_TIME)
        variant.time_value.FromDatetime(obj)
        self.estimated_pending_bytes += 16  # Header + size + (header + 32 bit number + header + 64 bit number)

    def _dump_list(self, collection, variant, current_depth, config, log_errors):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_LIST)

        if ListNamespace.is_numpy_obj(collection):
            collection = collection.tolist()
            if not collection:
                collection = []
        variant.original_size = len(collection)
        self.estimated_pending_bytes += 3  # Header + number

        # Dump only if we are not too deep
        if current_depth < config.max_collection_dump:

            for index, item in enumerate(collection):
                if index >= config.max_width:
                    break

                item_variant = variant.collection_values.add()
                self.estimated_pending_bytes += 3  # Header + size

                self._dump_python_object(item, item_variant, current_depth+1, config, log_errors)

    def _dump_dictionary(self, collection, variant, current_depth, config, log_errors):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_MAP)
        variant.original_size = len(collection)
        self.estimated_pending_bytes += 3  # Header + number

        # Dump only if we are not too deep
        if current_depth < config.max_collection_dump:

            i = 0

            for key, value in six.iteritems(collection):
                i += 1
                if i > config.max_width:
                    break

                key_variant = variant.collection_keys.add()
                value_variant = variant.collection_values.add()
                self.estimated_pending_bytes += 6  # Header + size + header + size

                self._dump_python_object(key, key_variant, current_depth+1, config, log_errors)
                self._dump_python_object(value, value_variant, current_depth+1, config, log_errors)

    def _dump_protobuf(self, obj, variant, current_depth, config, log_errors):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_OBJECT)
        if hasattr(obj, 'DESCRIPTOR'):
            for field in obj.ListFields():
                try:
                    variant.attribute_names_in_cache.append(self._get_string_index_in_cache(field[0].name))

                    attribute_value_variant = variant.attribute_values.add()
                    self.estimated_pending_bytes += 3  # Header + size

                    self._dump_python_object(field[1], attribute_value_variant, current_depth - 1, config, log_errors)
                except Exception:  # for now we just ignore errors when dumping protobuf
                    pass

    def _dump_exception(self, exc, variant, current_depth, config, log_errors):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_OBJECT)

        if len(exc.args) > 0:
            variant.attribute_names_in_cache.append(self._get_string_index_in_cache("args"))
            args_variant = variant.attribute_values.add()
            self.estimated_pending_bytes += 6  # Header + number + header + size

            self._dump_python_object(exc.args, args_variant, current_depth + 1, config, log_errors)

    def _dump_user_class(self, obj, variant, current_depth, config, log_errors):
        object_weight = current_depth + 1

        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_OBJECT)

        for key, value in six.iteritems(obj.__dict__.copy()):
            if key not in self.BUILTIN_ATTRIBUTES_IGNORE:
                if object_weight >= config.max_depth:
                    self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_OBJECT, True)
                    return

                variant.attribute_names_in_cache.append(self._get_string_index_in_cache(key))
                attribute_value_variant = variant.attribute_values.add()
                self.estimated_pending_bytes += 6  # Header + number + header + size

                self._dump_python_object(value, attribute_value_variant, object_weight, config, log_errors)

        if hasattr(obj, '__slots__') and obj.__slots__:
            items = obj.__slots__
            # py4j (used by pyspark to communicate with Java proxy objects) sets __slots__ to Java proxy objects,
            # and supports __dir__ instead.
            if not isinstance(items, Iterable):
                items = dir(items)
            for key in list(items):
                if key not in self.BUILTIN_ATTRIBUTES_IGNORE:
                    if object_weight >= config.max_depth:
                        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_OBJECT, True)
                        return

                    variant.attribute_names_in_cache.append(self._get_string_index_in_cache(key))
                    attribute_value_variant = variant.attribute_values.add()
                    self.estimated_pending_bytes += 6  # Header + number + header + size

                    try:
                        value = getattr(obj, key)
                    except AttributeError:
                        value = None
                    self._dump_python_object(value, attribute_value_variant, object_weight, config, log_errors)

    def _dump_not_supported(self, obj, variant):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_UKNOWN_OBJECT)

    def _dump_error_namespace(self, namespace, variant, log_errors):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_ERROR)
        variant.error_value.message = namespace.message.obj
        self.estimated_pending_bytes += 4 + len(namespace.message.obj)  # Large header + size + string size

        self.dump(namespace.parameters, variant.error_value.parameters, log_errors)
        self.dump(namespace.exc, variant.error_value.exc, log_errors)
        self.dump(namespace.traceback, variant.error_value.traceback, log_errors)

    def _dump_traceback_namespace(self, namespace, variant, log_errors):
        self.dump_variant_type(variant, variant_pb2.Variant.VARIANT_TRACEBACK)

        def increase_pending_size(size):
            self.estimated_pending_bytes += size

        namespace.dump(variant.code_values, self._get_string_index_in_cache, increase_pending_size)

    def dump_variant_type(self, variant, variant_type, max_depth=False):
        variant.variant_type_max_depth = (variant_type << 1) | int(max_depth)

        self.estimated_pending_bytes += 2  # Field header + short number

    def _get_bytes_index_in_cache(self, buffer):
        """
        Gets a buffer, store it in 'buffer_cache' if not stored yet and return its index

        @param buffer: The buffer
        @type buffer: bytes
        """

        if buffer in self.buffer_cache:
            return self.buffer_cache[buffer]

        current_size = len(self.buffer_cache)
        # Buffer length in bytes and an overhead of 5 bytes
        self.estimated_pending_bytes += len(buffer) + 5
        self.buffer_cache[buffer] = current_size
        return current_size

    def get_buffer_cache(self):
        return self.buffer_cache



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: envelope.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x65nvelope.proto\x12\x0b\x63om.rookout\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x19google/protobuf/any.proto\"\\\n\x08\x45nvelope\x12-\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12!\n\x03msg\x18\x02 \x01(\x0b\x32\x14.google.protobuf.AnyBO\n\x19\x63om.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2b\x06proto3')



_ENVELOPE = DESCRIPTOR.message_types_by_name['Envelope']
Envelope = _reflection.GeneratedProtocolMessageType('Envelope', (_message.Message,), {
  'DESCRIPTOR' : _ENVELOPE,
  '__module__' : 'envelope_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.Envelope)
  })
_sym_db.RegisterMessage(Envelope)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2'
  _ENVELOPE._serialized_start=91
  _ENVELOPE._serialized_end=183
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: variant.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rvariant.proto\x12\x0b\x63om.rookout\x1a\x1fgoogle/protobuf/timestamp.proto\"\x9c\x01\n\x05\x45rror\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12(\n\nparameters\x18\x03 \x01(\x0b\x32\x14.com.rookout.Variant\x12!\n\x03\x65xc\x18\x04 \x01(\x0b\x32\x14.com.rookout.Variant\x12\'\n\ttraceback\x18\x05 \x01(\x0b\x32\x14.com.rookout.Variant\"\xf5\x15\n\x07Variant\x12/\n\x0cvariant_type\x18\x01 \x01(\x0e\x32\x19.com.rookout.Variant.Type\x12\x15\n\roriginal_type\x18\x02 \x01(\t\x12\x33\n\nattributes\x18\x03 \x03(\x0b\x32\x1f.com.rookout.Variant.NamedValue\x12\x11\n\tmax_depth\x18\x04 \x01(\x08\x12$\n\x1coriginal_type_index_in_cache\x18\x05 \x01(\r\x12\x13\n\tint_value\x18\x0b \x01(\x05H\x00\x12\x14\n\nlong_value\x18\x0c \x01(\x03H\x00\x12\x16\n\x0c\x64ouble_value\x18\r \x01(\x01H\x00\x12\x33\n\x0c\x62inary_value\x18\x0e \x01(\x0b\x32\x1b.com.rookout.Variant.BinaryH\x00\x12\x33\n\x0cstring_value\x18\x0f \x01(\x0b\x32\x1b.com.rookout.Variant.StringH\x00\x12/\n\nlist_value\x18\x10 \x01(\x0b\x32\x19.com.rookout.Variant.ListH\x00\x12-\n\tmap_value\x18\x11 \x01(\x0b\x32\x18.com.rookout.Variant.MapH\x00\x12\x39\n\x0fnamespace_value\x18\x13 \x01(\x0b\x32\x1e.com.rookout.Variant.NamespaceH\x00\x12)\n\x0b\x65rror_value\x18\x14 \x01(\x0b\x32\x12.com.rookout.ErrorH\x00\x12>\n\rmessage_value\x18\x17 \x01(\x0b\x32%.com.rookout.Variant.FormattedMessageH\x00\x12\x30\n\ntime_value\x18\x18 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00\x12\x35\n\ncode_value\x18\x1a \x01(\x0b\x32\x1f.com.rookout.Variant.CodeObjectH\x00\x12\x38\n\x0flarge_int_value\x18\x1b \x01(\x0b\x32\x1d.com.rookout.Variant.LargeIntH\x00\x12\x35\n\rcomplex_value\x18\x1c \x01(\x0b\x32\x1c.com.rookout.Variant.ComplexH\x00\x12\x36\n\nenum_value\x18\x1f \x01(\x0b\x32 .com.rookout.Variant.EnumerationH\x00\x12\x33\n\ttraceback\x18  \x01(\x0b\x32\x1e.com.rookout.Variant.TracebackH\x00\x12\x38\n\x08livetail\x18! \x01(\x0b\x32$.com.rookout.Variant.LiveTailMessageH\x00\x1a?\n\nNamedValue\x12\x0c\n\x04name\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.com.rookout.Variant\x1a.\n\x06\x42inary\x12\x15\n\roriginal_size\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x0c\x1aL\n\x06String\x12\x15\n\roriginal_size\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\t\x12\x1c\n\x14value_index_in_cache\x18\x03 \x01(\r\x1aQ\n\x04List\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x15\n\roriginal_size\x18\x02 \x01(\x05\x12$\n\x06values\x18\x03 \x03(\x0b\x32\x14.com.rookout.Variant\x1aQ\n\x04Pair\x12#\n\x05\x66irst\x18\x01 \x01(\x0b\x32\x14.com.rookout.Variant\x12$\n\x06second\x18\x02 \x01(\x0b\x32\x14.com.rookout.Variant\x1a\x46\n\x03Map\x12\x15\n\roriginal_size\x18\x01 \x01(\x05\x12(\n\x05pairs\x18\x02 \x03(\x0b\x32\x19.com.rookout.Variant.Pair\x1a@\n\tNamespace\x12\x33\n\nattributes\x18\x01 \x03(\x0b\x32\x1f.com.rookout.Variant.NamedValue\x1a#\n\x10\x46ormattedMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\x1a\xa9\x01\n\nCodeObject\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06module\x18\x02 \x01(\t\x12\x10\n\x08\x66ilename\x18\x03 \x01(\t\x12\x0e\n\x06lineno\x18\x04 \x01(\r\x12\x1b\n\x13name_index_in_cache\x18\x05 \x01(\r\x12\x1d\n\x15module_index_in_cache\x18\x06 \x01(\r\x12\x1f\n\x17\x66ilename_index_in_cache\x18\x07 \x01(\r\x1a\x19\n\x08LargeInt\x12\r\n\x05value\x18\x01 \x01(\t\x1a*\n\x07\x43omplex\x12\x0c\n\x04real\x18\x01 \x01(\x01\x12\x11\n\timaginary\x18\x02 \x01(\x01\x1aM\n\x0b\x45numeration\x12\x14\n\x0cstring_value\x18\x01 \x01(\t\x12\x15\n\rordinal_value\x18\x02 \x01(\x05\x12\x11\n\ttype_name\x18\x03 \x01(\t\x1a?\n\tTraceback\x12\x32\n\tlocations\x18\x01 \x03(\x0b\x32\x1f.com.rookout.Variant.CodeObject\x1a\xc5\x02\n\x0fLiveTailMessage\x12\x12\n\nlevel_name\x18\x01 \x01(\t\x12\x0b\n\x03msg\x18\x02 \x01(\t\x12\x19\n\x11\x66ormatted_message\x18\x03 \x01(\t\x12\x0c\n\x04time\x18\x04 \x01(\x01\x12\x10\n\x08\x66ilename\x18\x05 \x01(\t\x12\x0e\n\x06lineno\x18\x06 \x01(\x05\x12\x10\n\x08\x66unction\x18\x07 \x01(\t\x12\x0e\n\x06module\x18\x08 \x01(\t\x12\x11\n\tthread_id\x18\t \x01(\x03\x12\x13\n\x0bthread_name\x18\n \x01(\t\x12I\n\x0blog_context\x18\x0b \x03(\x0b\x32\x34.com.rookout.Variant.LiveTailMessage.LogContextEntry\x1a\x31\n\x0fLogContextEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x97\x04\n\x04Type\x12\x10\n\x0cVARIANT_NONE\x10\x00\x12\x0f\n\x0bVARIANT_INT\x10\x01\x12\x10\n\x0cVARIANT_LONG\x10\x02\x12\x12\n\x0eVARIANT_DOUBLE\x10\x03\x12\x12\n\x0eVARIANT_BINARY\x10\x04\x12\x12\n\x0eVARIANT_STRING\x10\x05\x12\x10\n\x0cVARIANT_LIST\x10\x06\x12\x0f\n\x0bVARIANT_MAP\x10\x07\x12\x12\n\x0eVARIANT_OBJECT\x10\x08\x12\x15\n\x11VARIANT_NAMESPACE\x10\t\x12\x11\n\rVARIANT_ERROR\x10\n\x12\x19\n\x15VARIANT_NOT_SUPPORTED\x10\x0b\x12\x15\n\x11VARIANT_MAX_DEPTH\x10\x0c\x12\x1d\n\x19VARIANT_FORMATTED_MESSAGE\x10\r\x12\x10\n\x0cVARIANT_TIME\x10\x0e\x12\x19\n\x15VARIANT_UKNOWN_OBJECT\x10\x0f\x12\x17\n\x13VARIANT_CODE_OBJECT\x10\x10\x12\x15\n\x11VARIANT_LARGE_INT\x10\x11\x12\x13\n\x0fVARIANT_COMPLEX\x10\x12\x12\x15\n\x11VARIANT_UNDEFINED\x10\x13\x12\x13\n\x0fVARIANT_DYNAMIC\x10\x14\x12\x10\n\x0cVARIANT_ENUM\x10\x15\x12\x15\n\x11VARIANT_TRACEBACK\x10\x16\x12\x14\n\x10VARIANT_LIVETAIL\x10\x17\x12\x0f\n\x0bVARIANT_SET\x10\x18\x42\x07\n\x05valueBO\n\x19\x63om.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2b\x06proto3')



_ERROR = DESCRIPTOR.message_types_by_name['Error']
_VARIANT = DESCRIPTOR.message_types_by_name['Variant']
_VARIANT_NAMEDVALUE = _VARIANT.nested_types_by_name['NamedValue']
_VARIANT_BINARY = _VARIANT.nested_types_by_name['Binary']
_VARIANT_STRING = _VARIANT.nested_types_by_name['String']
_VARIANT_LIST = _VARIANT.nested_types_by_name['List']
_VARIANT_PAIR = _VARIANT.nested_types_by_name['Pair']
_VARIANT_MAP = _VARIANT.nested_types_by_name['Map']
_VARIANT_NAMESPACE = _VARIANT.nested_types_by_name['Namespace']
_VARIANT_FORMATTEDMESSAGE = _VARIANT.nested_types_by_name['FormattedMessage']
_VARIANT_CODEOBJECT = _VARIANT.nested_types_by_name['CodeObject']
_VARIANT_LARGEINT = _VARIANT.nested_types_by_name['LargeInt']
_VARIANT_COMPLEX = _VARIANT.nested_types_by_name['Complex']
_VARIANT_ENUMERATION = _VARIANT.nested_types_by_name['Enumeration']
_VARIANT_TRACEBACK = _VARIANT.nested_types_by_name['Traceback']
_VARIANT_LIVETAILMESSAGE = _VARIANT.nested_types_by_name['LiveTailMessage']
_VARIANT_LIVETAILMESSAGE_LOGCONTEXTENTRY = _VARIANT_LIVETAILMESSAGE.nested_types_by_name['LogContextEntry']
_VARIANT_TYPE = _VARIANT.enum_types_by_name['Type']
Error = _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), {
  'DESCRIPTOR' : _ERROR,
  '__module__' : 'variant_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.Error)
  })
_sym_db.RegisterMessage(Error)

Variant = _reflection.GeneratedProtocolMessageType('Variant', (_message.Message,), {

  'NamedValue' : _reflection.GeneratedProtocolMessageType('NamedValue', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_NAMEDVALUE,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.NamedValue)
    })
  ,

  'Binary' : _reflection.GeneratedProtocolMessageType('Binary', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_BINARY,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.Binary)
    })
  ,

  'String' : _reflection.GeneratedProtocolMessageType('String', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_STRING,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.String)
    })
  ,

  'List' : _reflection.GeneratedProtocolMessageType('List', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_LIST,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.List)
    })
  ,

  'Pair' : _reflection.GeneratedProtocolMessageType('Pair', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_PAIR,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.Pair)
    })
  ,

  'Map' : _reflection.GeneratedProtocolMessageType('Map', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_MAP,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.Map)
    })
  ,

  'Namespace' : _reflection.GeneratedProtocolMessageType('Namespace', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_NAMESPACE,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.Namespace)
    })
  ,

  'FormattedMessage' : _reflection.GeneratedProtocolMessageType('FormattedMessage', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_FORMATTEDMESSAGE,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.FormattedMessage)
    })
  ,

  'CodeObject' : _reflection.GeneratedProtocolMessageType('CodeObject', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_CODEOBJECT,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.CodeObject)
    })
  ,

  'LargeInt' : _reflection.GeneratedProtocolMessageType('LargeInt', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_LARGEINT,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.LargeInt)
    })
  ,

  'Complex' : _reflection.GeneratedProtocolMessageType('Complex', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_COMPLEX,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.Complex)
    })
  ,

  'Enumeration' : _reflection.GeneratedProtocolMessageType('Enumeration', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_ENUMERATION,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.Enumeration)
    })
  ,

  'Traceback' : _reflection.GeneratedProtocolMessageType('Traceback', (_message.Message,), {
    'DESCRIPTOR' : _VARIANT_TRACEBACK,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.Traceback)
    })
  ,

  'LiveTailMessage' : _reflection.GeneratedProtocolMessageType('LiveTailMessage', (_message.Message,), {

    'LogContextEntry' : _reflection.GeneratedProtocolMessageType('LogContextEntry', (_message.Message,), {
      'DESCRIPTOR' : _VARIANT_LIVETAILMESSAGE_LOGCONTEXTENTRY,
      '__module__' : 'variant_pb2'
      # @@protoc_insertion_point(class_scope:com.rookout.Variant.LiveTailMessage.LogContextEntry)
      })
    ,
    'DESCRIPTOR' : _VARIANT_LIVETAILMESSAGE,
    '__module__' : 'variant_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.Variant.LiveTailMessage)
    })
  ,
  'DESCRIPTOR' : _VARIANT,
  '__module__' : 'variant_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.Variant)
  })
_sym_db.RegisterMessage(Variant)
_sym_db.RegisterMessage(Variant.NamedValue)
_sym_db.RegisterMessage(Variant.Binary)
_sym_db.RegisterMessage(Variant.String)
_sym_db.RegisterMessage(Variant.List)
_sym_db.RegisterMessage(Variant.Pair)
_sym_db.RegisterMessage(Variant.Map)
_sym_db.RegisterMessage(Variant.Namespace)
_sym_db.RegisterMessage(Variant.FormattedMessage)
_sym_db.RegisterMessage(Variant.CodeObject)
_sym_db.RegisterMessage(Variant.LargeInt)
_sym_db.RegisterMessage(Variant.Complex)
_sym_db.RegisterMessage(Variant.Enumeration)
_sym_db.RegisterMessage(Variant.Traceback)
_sym_db.RegisterMessage(Variant.LiveTailMessage)
_sym_db.RegisterMessage(Variant.LiveTailMessage.LogContextEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2'
  _VARIANT_LIVETAILMESSAGE_LOGCONTEXTENTRY._options = None
  _VARIANT_LIVETAILMESSAGE_LOGCONTEXTENTRY._serialized_options = b'8\001'
  _ERROR._serialized_start=64
  _ERROR._serialized_end=220
  _VARIANT._serialized_start=223
  _VARIANT._serialized_end=3028
  _VARIANT_NAMEDVALUE._serialized_start=1236
  _VARIANT_NAMEDVALUE._serialized_end=1299
  _VARIANT_BINARY._serialized_start=1301
  _VARIANT_BINARY._serialized_end=1347
  _VARIANT_STRING._serialized_start=1349
  _VARIANT_STRING._serialized_end=1425
  _VARIANT_LIST._serialized_start=1427
  _VARIANT_LIST._serialized_end=1508
  _VARIANT_PAIR._serialized_start=1510
  _VARIANT_PAIR._serialized_end=1591
  _VARIANT_MAP._serialized_start=1593
  _VARIANT_MAP._serialized_end=1663
  _VARIANT_NAMESPACE._serialized_start=1665
  _VARIANT_NAMESPACE._serialized_end=1729
  _VARIANT_FORMATTEDMESSAGE._serialized_start=1731
  _VARIANT_FORMATTEDMESSAGE._serialized_end=1766
  _VARIANT_CODEOBJECT._serialized_start=1769
  _VARIANT_CODEOBJECT._serialized_end=1938
  _VARIANT_LARGEINT._serialized_start=1940
  _VARIANT_LARGEINT._serialized_end=1965
  _VARIANT_COMPLEX._serialized_start=1967
  _VARIANT_COMPLEX._serialized_end=2009
  _VARIANT_ENUMERATION._serialized_start=2011
  _VARIANT_ENUMERATION._serialized_end=2088
  _VARIANT_TRACEBACK._serialized_start=2090
  _VARIANT_TRACEBACK._serialized_end=2153
  _VARIANT_LIVETAILMESSAGE._serialized_start=2156
  _VARIANT_LIVETAILMESSAGE._serialized_end=2481
  _VARIANT_LIVETAILMESSAGE_LOGCONTEXTENTRY._serialized_start=2432
  _VARIANT_LIVETAILMESSAGE_LOGCONTEXTENTRY._serialized_end=2481
  _VARIANT_TYPE._serialized_start=2484
  _VARIANT_TYPE._serialized_end=3019
# @@protoc_insertion_point(module_scope)



import socket
import errno
import select
import websocket
import threading
import time

from rook.com_ws.agent_com_impel.agent_com_base import AgentComBase, FlushMessagesEvent
from rook.com_ws.envelope_wrappers.basic_envelope_wrapper import BasicEnvelopeWrapper
from rook.config import AgentComConfiguration

from rook.logger import logger
from six.moves.queue import Queue, Empty
import rook.protobuf.envelope_pb2 as envelope_pb


class ExitThreadSentinel(object):
    pass


class MultiThreadAgentCom(AgentComBase):

    def __init__(self, agent_id, host, port, proxy, token, labels, tags, debug, print_on_initial_connection):
        super(MultiThreadAgentCom, self).__init__(agent_id, host, port, proxy, token, labels, tags, debug,
                                                  print_on_initial_connection)

        # Queue hold outgoing messages wrapped with IEnvelopeWrapper
        self._queue = Queue()

    def await_message(self, message_name):
        event = threading.Event()
        self.once(message_name, lambda _: event.set())

        return event

    def _create_run_connection_thread(self):
        stop_socket_incoming, incoming_client_socket = socket.socketpair()
        outgoing_exit_sentinel = ExitThreadSentinel()

        def signal_stop_incoming_thread():
            # Logging is not safe here, interpreter might be shutting down
            try:
                stop_socket_incoming.send(b'1')
            except socket.error as socket_error:
                if socket_error.errno == errno.EPIPE:
                    return

                logger.debug("Failed to signal stop", exc_info=1)
            except Exception:
                logger.debug("Failed to signal stop", exc_info=1)
            finally:
                try:
                    stop_socket_incoming.close()
                except:  # lgtm[py/catch-base-exception]
                    pass

        def signal_stop_outgoing_thread():
            # Logging is not safe here, interpreter might be shutting down
            try:
                incoming_client_socket.close()
            except:  # Nothing to do as we are already closing
                pass

            try:
                self._queue.put(BasicEnvelopeWrapper(outgoing_exit_sentinel))
            except Exception:
                logger.debug("Failed to signal stop", exc_info=1)

        routines = [threading.Thread(name="rookout-incoming-thread", target=self._incoming,
                                     args=(incoming_client_socket, signal_stop_outgoing_thread)),
                    threading.Thread(name="rookout-outgoing-thread", target=self._outgoing,
                                     args=(outgoing_exit_sentinel, signal_stop_incoming_thread))]
        for routine in routines:
            routine.daemon = True
            routine.start()
        logger.info("Finished initialization")
        for routine in routines:
            routine.join()

    def _outgoing(self, outgoing_exit_sentinel, on_exit):
        try:
            last_ping = time.time()
            self._connection.ping()

            while self._running:
                envelope_wrapper = None
                if (time.time() - last_ping) >= AgentComConfiguration.PING_INTERVAL:
                    last_ping = time.time()
                    self._connection.ping()

                try:
                    envelope_wrapper = self._queue.get(timeout=AgentComConfiguration.PING_INTERVAL)
                    if envelope_wrapper is None:
                        break
                    msg = envelope_wrapper.get_buffer()
                    if isinstance(msg, ExitThreadSentinel):
                        if msg is outgoing_exit_sentinel:
                            break
                        continue  # if it's an ExitThreadSentinel but not from this specific thread, just skip it
                    if isinstance(msg, FlushMessagesEvent):
                        msg.event.set()
                        continue
                    self._send(msg)
                except Empty:
                    continue
                except (socket.error, websocket.WebSocketConnectionClosedException):
                    if envelope_wrapper is not None:
                        self._queue.put(envelope_wrapper)
                    break

                if envelope_wrapper is not None:
                    self._queue_messages_length -= len(envelope_wrapper)

        except:  # lgtm[py/catch-base-exception]
            try:
                logger.exception("Outgoing thread failed")
            except:  # lgtm[py/catch-base-exception]
                pass
        finally:
            on_exit()

    def _incoming(self, stop_socket, on_exit):
        try:
            while self._running:
                try:
                    # Poll has no file_descriptor limit (on linux) if available - use it
                    if self.poll_available:
                        if self.wait_for_event_poll(stop_socket):
                            break
                    else:
                        if self.wait_for_event_select(stop_socket):
                            break

                    # it wasn't stop_socket, so it's self._connection -> we can read. we have to read including
                    # control frames, otherwise select might return and recv() wouldn't actually return anything
                    code, msg = self._connection.recv_data(control_frame=True)
                    if code != websocket.ABNF.OPCODE_BINARY:
                        continue

                    if msg is None:
                        # socket disconnected
                        logger.debug("Incoming thread - socket disconnected")
                        break

                    if len(msg) > AgentComConfiguration.AGENT_COM_INCOMING_MAX_MESSAGE_SIZE:
                        logger.error("message length (%d) exceed max size", msg)
                        continue

                    envelope = envelope_pb.Envelope()
                    envelope.ParseFromString(msg)
                    self._handle_incoming_message(envelope)
                except (socket.error, websocket.WebSocketConnectionClosedException):
                    logger.debug("Incoming thread - socket disconnected")
                    break
        except:  # lgtm[py/catch-base-exception]
            try:
                logger.exception("Incoming thread failed")
            except:  # lgtm[py/catch-base-exception]
                pass
        finally:
            on_exit()

    def wait_for_event_poll(self, stop_socket):
        poller = select.poll()

        exception_flags = select.POLLHUP | select.POLLERR
        read_only_flags = select.POLLIN | select.POLLPRI | exception_flags

        # Registering all input close and exception types
        poller.register(self._connection, read_only_flags)
        poller.register(stop_socket, read_only_flags)

        events = poller.poll(AgentComConfiguration.PING_TIMEOUT * 1000)

        try:
            for fd, flag in events:
                if fd == stop_socket.fileno():
                    return True  # signaled to stop
                elif fd == self._connection.fileno() and flag \
                        and flag & exception_flags:  # Checking these flags explicitly (ignore input data)
                    return True
        except AttributeError:  # happens when the socket is already None (and fileno() will fail)
            return True  # Should stop

        if len(events) == 0:
            logger.debug("Incoming thread - ping timeout")
            return True  # select timed out - ping timeout

        # Every thing is ok
        return False

    def wait_for_event_select(self, stop_socket):
        rlist, _, xlist = select.select([self._connection, stop_socket],
                                        [],
                                        [self._connection, stop_socket],
                                        AgentComConfiguration.PING_TIMEOUT)
        if stop_socket in rlist or stop_socket in xlist or self._connection in xlist:
            return True  # signaled to stop

        if len(rlist) == 0 and len(xlist) == 0:
            logger.debug("Incoming thread - ping timeout")
            return True  # select timed out - ping timeout
        return False



import datetime
import traceback
import decimal
from types import TracebackType
from six.moves.collections_abc import Iterable

import six

from google.protobuf.internal.type_checkers import Int32ValueChecker, Int64ValueChecker
from rook.processor.namespace_serializer_base import NamespaceSerializerBase

from .namespaces.container_namespace import ContainerNamespace
from .namespaces.python_object_namespace import PythonObjectNamespace
from .namespaces.collection_namespace import DictNamespace, ListNamespace, LIST_TYPE
from .namespaces.traceback_namespace import TracebackNamespace
from .namespaces.user_object_namespace import UserObjectNamespace
from .namespaces.error_namespace import ErrorNamespace
from .namespaces.max_depth_namespace import MaxDepthNamespace
from .namespaces.formatted_namespace import FormattedNamespace
from .namespaces.dumped_primitive_namespace import DumpedPrimitiveNamespace
from .namespaces.string_namespace import StringNamespace
from .namespaces.unknown_namespace import UnknownNamespace
from .namespaces.code_object_namespace import CodeObjectNamespace
from .namespaces.dynamic_object_namespace import DynamicObjectNamespace

from rook.logger import logger

from rook.protobuf import variant_pb2


class NamespaceSerializer(NamespaceSerializerBase):
    def __init__(self, use_string_cache=False):
        NamespaceSerializerBase.__init__(self, use_string_cache)

    def dump(self, namespace, variant, log_errors=True):
        try:
            if isinstance(namespace, ContainerNamespace):
                self._dump_container_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, PythonObjectNamespace):
                self._dump_object_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, StringNamespace):
                self._dump_string_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, DumpedPrimitiveNamespace):
                self._dump_primitive_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, UserObjectNamespace):
                self._dump_user_object_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, ListNamespace):
                self._dump_list_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, DictNamespace):
                self._dump_dictionary_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, ErrorNamespace):
                self._dump_error_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, MaxDepthNamespace):
                self._dump_max_depth_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, FormattedNamespace):
                self._dump_formatted_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, UnknownNamespace):
                self._dump_unknown_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, TracebackNamespace):
                self._dump_traceback_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, CodeObjectNamespace):
                self.dump_code_namespace(namespace, variant, log_errors)
            elif isinstance(namespace, DynamicObjectNamespace):
                self.dump_dynamic_namespace(namespace, variant, log_errors)
            else:
                raise NotImplementedError("Does not support serializing this type!", type(namespace))
        except Exception as e:
            message = "Failed to serialize namespace"

            from .error import Error  # lgtm[py/cyclic-import]
            variant.Clear()
            variant.variant_type = variant.VARIANT_ERROR

            if log_errors:
                Error(exc=e, message=message).dump(variant.error_value)
                logger.exception(message)

    def dumps(self, namespace, log_errors=True):
        variant = variant_pb2.Variant()
        self.dump(namespace, variant, log_errors)
        return variant

    # TODO - remove this entire flow (used only for testing, barely even that)
    def loads(self, message):
        return self._load_variant(message)

    def _dump_container_namespace(self, namespace, variant, log_errors):
        variant.variant_type = variant.VARIANT_NAMESPACE

        for key, value in six.iteritems(namespace.dictionary):
            item = variant.namespace_value.attributes.add()
            item.name = key
            self.dump(value, item.value, log_errors)

    def _dump_object_namespace(self, namespace, variant, log_errors):
        self._dump_python_object(namespace.obj, variant, 0, namespace.dump_config, log_errors)

    def _dump_string_namespace(self, namespace, variant, log_errors):
        self._dump_dumped_object_namespace(namespace, variant, log_errors)

        if "string" == namespace.common_type:
            variant.variant_type = variant.VARIANT_STRING
            variant.string_value.original_size = namespace.original_size
            variant.string_value.value = namespace.obj
        elif "binary" == namespace.common_type:
            variant.variant_type = variant.VARIANT_BINARY
            variant.binary_value.original_size = namespace.original_size
            variant.binary_value.value = namespace.obj
        else:
            raise ValueError("Invalid string type %s", str(type(namespace.obj)))

    def _dump_primitive_namespace(self, namespace, variant, log_errors):
        self._dump_dumped_object_namespace(namespace, variant, log_errors)
        self._dump_primitive(namespace.obj, namespace.type, variant, PythonObjectNamespace.ObjectDumpConfig.UNLIMITED_STRING)

    def _dump_formatted_namespace(self, namespace, variant, log_errors):
        variant.variant_type = variant.VARIANT_FORMATTED_MESSAGE
        variant.message_value.message = namespace.obj

    def _dump_unknown_namespace(self, namespace, variant, log_errors):
        variant.variant_type = variant.VARIANT_UKNOWN_OBJECT
        variant.original_type = namespace.type

    def _dump_python_object(self, obj, variant, current_depth, config, log_errors):
        try:
            self._dump_python_object_unsafe(obj, variant, current_depth, config, log_errors)
        except Exception as e:
            message = "Failed to serialize namespace"

            variant.Clear()
            variant.variant_type = variant.VARIANT_ERROR

            if log_errors:
                from .error import Error
                Error(exc=e, message=message).dump(variant.error_value)
                logger.exception(message)

    def _dump_python_object_unsafe(self, obj, variant, current_depth, config, log_errors):
        self._dump_base_object(obj, variant, current_depth, config, log_errors)
        if isinstance(obj, NamespaceSerializerBase.PRIMITIVE_TYPES):
            self._dump_primitive(obj, str(type(obj)), variant, config.max_string)
        elif isinstance(obj, LIST_TYPE):
            self._dump_list(obj, variant, current_depth, config, log_errors)
        elif isinstance(obj, dict):
            self._dump_dictionary(obj, variant, current_depth, config, log_errors)
        elif isinstance(obj, BaseException):
            self._dump_exception(obj, variant, current_depth, config, log_errors)
        elif isinstance(obj, TracebackType):
            self._dump_traceback(obj, variant, current_depth, config, log_errors)
        elif NamespaceSerializer.is_numpy_obj(obj):
            self._dump_primitive(obj.item(), str(type(obj)), variant, config.max_string)
        elif NamespaceSerializer.is_torch_obj(obj):
            self._dump_primitive(str(obj), str(type(obj)), variant, config.max_string)
        elif NamespaceSerializer.is_multidict_obj(obj):
            self._dump_primitive(str(obj), str(type(obj)), variant, config.max_string)
        elif NamespaceSerializer.is_protobuf_obj(obj):
            self._dump_protobuf(obj, variant, current_depth, config, log_errors)
        elif hasattr(obj, '__dict__'):
            self._dump_user_class(obj, variant, current_depth, config)
        else:
            self._dump_not_supported(obj, variant)

    def _dump_traceback(self, obj, variant, current_depth, config, log_errors):
        # python separates the "forward" stack (callees of the except clause)
        # and the "backward" stack (callers of above)
        # Possibly would be more useful to wrap this in a StackNamespace
        tb = traceback.format_tb(obj)
        tb[1:1] = traceback.format_stack(obj.tb_frame.f_back)
        value = ''.join(tb)

        variant.variant_type = variant.VARIANT_STRING
        variant.string_value.original_size = len(value)
        variant.string_value.value = value

    def _dump_base_object(self, obj, variant, current_depth, config, log_errors):
        original_type = str(type(obj))
        if not self.use_string_cache:
            variant.original_type = original_type
        else:
            variant.original_type_index_in_cache = self._get_string_index_in_cache(original_type)

        object_width = self._get_object_width(obj)
        object_weight = current_depth + (2 if object_width >= config.max_width else 1)
        if object_weight >= config.max_depth:
            if object_width:
                self._set_max_depth_flag(variant)
            return

        if hasattr(obj, '__dict__'):
            # __dict is copied before iteration, as it might
            # change in another thread. We do not deep copy,
            # because we don't really mind if the values
            # change.
            for key, value in six.iteritems(obj.__dict__.copy()):
                if key not in self.BUILTIN_ATTRIBUTES_IGNORE:
                    attribute = variant.attributes.add()
                    attribute.name = key
                    self._dump_python_object(value, attribute.value, object_weight, config, log_errors)

        if hasattr(obj, '__slots__') and obj.__slots__:
            items = obj.__slots__
            # py4j (used by pyspark to communicate with Java proxy objects) sets __slots__ to Java proxy objects,
            # and supports __dir__ instead.
            if not isinstance(items, Iterable):
                items = dir(items)
            for key in list(items):
                if key not in self.BUILTIN_ATTRIBUTES_IGNORE:
                    attribute = variant.attributes.add()
                    attribute.name = key
                    try:
                        value = getattr(obj, key)
                    except AttributeError:
                        value = None
                    self._dump_python_object(value, attribute.value, object_weight, config, log_errors)

    def _dump_max_depth(self, variant):
        variant.variant_type = variant.VARIANT_MAX_DEPTH

    def _set_max_depth_flag(self, variant):
        variant.max_depth = True

    def _dump_primitive(self, obj, original_type, variant, max_string):
        if obj is None:
            variant.variant_type = variant.VARIANT_NONE

        elif isinstance(obj, int) and (obj < Int32ValueChecker._MAX) and (obj > Int32ValueChecker._MIN):
            variant.variant_type = variant.VARIANT_INT
            variant.int_value = int(obj)

        elif isinstance(obj, six.integer_types) and (obj < Int64ValueChecker._MAX) and (obj > Int64ValueChecker._MIN):
            variant.variant_type = variant.VARIANT_LONG
            variant.long_value = int(obj)

        elif isinstance(obj, six.integer_types):
            variant.variant_type = variant.VARIANT_LARGE_INT
            variant.large_int_value.value = str(obj)

        elif isinstance(obj, bool):
            variant.variant_type = variant.VARIANT_INT
            variant.int_value = int(obj)

        elif isinstance(obj, float):
            variant.variant_type = variant.VARIANT_DOUBLE
            variant.double_value = float(obj)

        elif isinstance(obj, decimal.Decimal):
            serialized_decimal = str(obj)

            variant.variant_type = variant.VARIANT_STRING
            variant.string_value.original_size = len(serialized_decimal)
            variant.string_value.value = serialized_decimal

        elif isinstance(obj, six.string_types):
            variant.variant_type = variant.VARIANT_STRING
            variant.string_value.original_size = len(obj)

            if len(obj) > max_string:
                obj = obj[:max_string]

            string = self.normalize_string(obj)
            if not self.use_string_cache:
                variant.string_value.value = string # This is legacy code.
            else:
                variant.string_value.value_index_in_cache = self._get_string_index_in_cache(string)

        elif isinstance(obj, self.BINARY_TYPES) or original_type == 'binary_type':
            variant.variant_type = variant.VARIANT_BINARY
            variant.binary_value.original_size = len(obj)

            if len(obj) > max_string:
                obj = obj[:max_string]

            variant.binary_value.value = bytes(obj)

        elif isinstance(obj, self.CODE_TYPES):
            variant.variant_type = variant.VARIANT_CODE_OBJECT
            variant.code_value.name = self.normalize_string(obj.__name__)
            if hasattr(obj, '__code__') and hasattr(obj.__code__, 'co_filename'):
                variant.code_value.filename = self.normalize_string(obj.__code__.co_filename)
                variant.code_value.lineno = int(obj.__code__.co_firstlineno)
            if hasattr(obj, '__module__') and obj.__module__:
                variant.code_value.module = self.normalize_string(obj.__module__)

        elif isinstance(obj, complex):
            variant.variant_type = variant.VARIANT_COMPLEX
            variant.complex_value.real = float(obj.real)
            variant.complex_value.imaginary = float(obj.imag)

        elif isinstance(obj, datetime.datetime):
            self._dump_datetime(obj, variant)

        else:
            raise ValueError("Object is not a supported primitive!", type(obj))

    def _dump_datetime(self, obj, variant):
        if obj.tzinfo:
            obj = obj.replace(tzinfo=None)

        variant.variant_type = variant.VARIANT_TIME
        variant.time_value.FromDatetime(obj)

    def _dump_list(self, collection, variant, current_depth, config, log_errors):
        variant.variant_type = variant.VARIANT_LIST
        variant.list_value.type = ListNamespace.get_common_type(collection)
        if ListNamespace.is_numpy_obj(collection):
            collection = collection.tolist()
            if not collection:
                collection = []
        variant.list_value.original_size = len(collection)

        # Dump only if we are not too deep
        if current_depth < config.max_collection_dump:

            for index, item in enumerate(collection):
                if index >= config.max_width:
                    break

                variant_item = variant.list_value.values.add()
                self._dump_python_object(item, variant_item, current_depth+1, config, log_errors)

    def _dump_dictionary(self, collection, variant, current_depth, config, log_errors):
        variant.variant_type = variant.VARIANT_MAP
        variant.map_value.original_size = len(collection)

        # Dump only if we are not too deep
        if current_depth < config.max_collection_dump:

            i = 0

            for key, value in six.iteritems(collection):
                i += 1
                if i > config.max_width:
                    break

                pair = variant.map_value.pairs.add()
                self._dump_python_object(key, pair.first, current_depth+1, config, log_errors)
                self._dump_python_object(value, pair.second, current_depth+1, config, log_errors)

    def _dump_protobuf(self, obj, variant, current_depth, config, log_errors):
        variant.variant_type = variant.VARIANT_OBJECT
        if hasattr(obj, 'DESCRIPTOR'):
            for field in obj.ListFields():
                try:
                    attribute = variant.attributes.add()
                    attribute.name = field[0].name

                    self._dump_python_object(field[1], attribute.value, current_depth - 1, config,
                                                 log_errors)
                except Exception:  # for now we just ignore errors when dumping protobuf
                    pass

    def _dump_exception(self, exc, variant, current_depth, config, log_errors):
        variant.variant_type = variant.VARIANT_OBJECT

        args = variant.attributes.add()
        args.name = "args"
        self._dump_python_object(exc.args, args.value, current_depth + 1, config, log_errors)

    def _dump_user_class(self, obj, variant, current_depth, config):
        variant.variant_type = variant.VARIANT_OBJECT

    def _dump_not_supported(self, obj, variant):
        variant.variant_type = variant.VARIANT_UKNOWN_OBJECT

    def _dump_dumped_object_namespace(self, namespace, variant, log_errors):
        variant.original_type = namespace.type

        for key, value in six.iteritems(namespace.attributes):
            attribute = variant.attributes.add()
            attribute.name = key
            self.dump(value, attribute.value, log_errors)

    def _dump_user_object_namespace(self, namespace, variant, log_errors):
        self._dump_dumped_object_namespace(namespace, variant, log_errors)
        variant.variant_type = variant.VARIANT_OBJECT

    def _dump_list_namespace(self, namespace, variant, log_errors):
        self._dump_dumped_object_namespace(namespace, variant, log_errors)

        variant.variant_type = variant.VARIANT_LIST
        variant.list_value.type = namespace.common_type
        variant.list_value.original_size = namespace.original_size

        for item in namespace:
            variant_item = variant.list_value.values.add()
            self.dump(item, variant_item, log_errors)

    def _dump_dictionary_namespace(self, namespace, variant, log_errors):
        self._dump_dumped_object_namespace(namespace, variant, log_errors)

        variant.variant_type = variant.VARIANT_MAP
        variant.map_value.original_size = namespace.original_size

        for key, value in six.iteritems(namespace):
            pair = variant.map_value.pairs.add()

            self.dump(key, pair.first, log_errors)
            self.dump(value, pair.second, log_errors)

    def _dump_error_namespace(self, namespace, variant, log_errors):
        variant.variant_type = variant.VARIANT_ERROR
        variant.error_value.message = namespace.message.obj
        self.dump(namespace.parameters, variant.error_value.parameters, log_errors)
        self.dump(namespace.exc, variant.error_value.exc, log_errors)
        self.dump(namespace.traceback, variant.error_value.traceback, log_errors)

    def _dump_max_depth_namespace(self, namespace, variant, log_errors):
        variant.variant_type = variant.VARIANT_MAX_DEPTH

    def _dump_traceback_namespace(self, namespace, variant, log_errors):
        variant.variant_type = variant_pb2.Variant.VARIANT_TRACEBACK

        if not self.use_string_cache:
            namespace.dump(variant.traceback.locations)
        else:
            namespace.dump(variant.traceback.locations, self._get_string_index_in_cache)

    def dump_code_namespace(self, namespace, variant, log_errors):
        self._dump_dumped_object_namespace(namespace, variant, log_errors)

        variant.variant_type = variant.VARIANT_CODE_OBJECT
        variant.code_value.name = namespace.name
        variant.code_value.module = namespace.module
        variant.code_value.filename = namespace.filename
        variant.code_value.lineno = namespace.lineno

    def dump_dynamic_namespace(self, namespace, variant, log_errors):
        variant.variant_type = variant.VARIANT_DYNAMIC

    def dump_type_namespace(self, namespace, variant, log_errors):
        self._dump_dumped_object_namespace(namespace, variant, log_errors)

        variant.variant_type = variant.VARIANT_TYPE
        variant.type_value.name = namespace.name

    def _load_variant(self, variant):
        # NOTE: This the common types are partially duplicated with PythonObjectNamespace.get_common_type

        try:
            if variant.variant_type == variant_pb2.Variant.VARIANT_NONE:
                return DumpedPrimitiveNamespace(None, variant.original_type, "null", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_INT:
                return DumpedPrimitiveNamespace(variant.int_value, variant.original_type, "int", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_LONG:
                return DumpedPrimitiveNamespace(variant.long_value, variant.original_type, "int", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_DOUBLE:
                return DumpedPrimitiveNamespace(variant.double_value, variant.original_type, "float", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_BINARY:
                return StringNamespace(variant.binary_value.value, variant.binary_value.original_size,
                                       variant.original_type, "binary", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_STRING:
                if variant.original_type == str(decimal.Decimal):
                    return DumpedPrimitiveNamespace(
                        decimal.Decimal(variant.string_value.value),
                        variant.original_type, "float", self._load_attributes(variant))
                return StringNamespace(variant.string_value.value, variant.string_value.original_size,
                                       variant.original_type, "string", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_TIME:
                return DumpedPrimitiveNamespace(variant.time_value.ToDatetime(), variant.original_type,
                                                "datetime", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_LIST:
                return self._load_list(variant)
            elif variant.variant_type == variant_pb2.Variant.VARIANT_MAP:
                return self._load_dictionary(variant)
            elif variant.variant_type == variant_pb2.Variant.VARIANT_OBJECT:
                return self._load_user_class(variant)
            elif variant.variant_type == variant_pb2.Variant.VARIANT_NAMESPACE:
                return self._load_container_namespace(variant)
            elif variant.variant_type == variant_pb2.Variant.VARIANT_ERROR:
                return self._load_error(variant.error_value)
            elif variant.variant_type == variant_pb2.Variant.VARIANT_MAX_DEPTH:
                return self._load_max_depth()
            elif variant.variant_type == variant_pb2.Variant.VARIANT_FORMATTED_MESSAGE:
                return self._load_formatted_namespace(variant)
            elif variant.variant_type == variant_pb2.Variant.VARIANT_UKNOWN_OBJECT:
                return UnknownNamespace(variant.original_type, self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_CODE_OBJECT:
                return self._load_code_object(variant)
            elif variant.variant_type == variant_pb2.Variant.VARIANT_LARGE_INT:
                return DumpedPrimitiveNamespace(int(variant.large_int_value.value), variant.original_type,
                                                "int", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_COMPLEX:
                return DumpedPrimitiveNamespace(complex(variant.complex_value.real, variant.complex_value.imaginary),
                                                variant.original_type, "complex", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_UNDEFINED:
                return DumpedPrimitiveNamespace(None, variant.original_type, "null", self._load_attributes(variant))
            elif variant.variant_type == variant_pb2.Variant.VARIANT_DYNAMIC:
                return DynamicObjectNamespace()
            else:
                raise ValueError("Invalid type", variant.variant_type)
        except Exception as e:
            message = "Failed to serialize namespace"
            logger.exception(message)

            from .error import Error  # lgtm[py/cyclic-import]
            error = Error(exc=e, message=message)

            # This is a duplication of Error as we can't use it here safely
            return ErrorNamespace(PythonObjectNamespace(message),
                                  PythonObjectNamespace(error.parameters),
                                  PythonObjectNamespace(error.exc),
                                  PythonObjectNamespace('\n'.join(traceback.format_tb(error.traceback))))

    def _load_attributes(self, variant):
        attributes = dict()

        for attribute in variant.attributes:
            attributes[attribute.name] = self._load_variant(attribute.value)

        return attributes

    def _load_list(self, variant):
        result = list()

        for item in variant.list_value.values:
            result.append(self._load_variant(item))

        return ListNamespace(result, variant.list_value.original_size, variant.original_type, variant.list_value.type,
                             self._load_attributes(variant))

    def _load_dictionary(self, variant):
        result = dict()

        for item in variant.map_value.pairs:
            result[self._load_variant(item.first)] = self._load_variant(item.second)

        return DictNamespace(result, variant.map_value.original_size, variant.original_type,
                             self._load_attributes(variant))

    def _load_user_class(self, variant):
        return UserObjectNamespace(variant.original_type, self._load_attributes(variant))

    def _load_container_namespace(self, variant):
        result = dict()

        for attribute in variant.namespace_value.attributes:
            result[attribute.name] = self._load_variant(attribute.value)

        return ContainerNamespace(result)

    def _load_error(self, error):
        return ErrorNamespace(PythonObjectNamespace(error.message),
                              self._load_variant(error.parameters),
                              self._load_variant(error.exc),
                              self._load_variant(error.traceback))

    def _load_max_depth(self):
        return MaxDepthNamespace()

    def _load_formatted_namespace(self, variant):
        return FormattedNamespace(variant.message_value.message)

    def _load_code_object(self, variant):
        return CodeObjectNamespace(variant.code_value.name,
                                   variant.code_value.module,
                                   variant.code_value.filename,
                                   variant.code_value.lineno,
                                   variant.original_type,
                                   self._load_attributes(variant))



import sys
import platform
from os import environ
import inspect

PLATFORMS_WITH_FORK_SUPPORT = ('darwin', 'linux2', 'linux')
PYTHON_IMPLEMENTATION_WITH_FORK_SUPPORT = 'CPython'
# Only works with Gunicorn 0.11.2 and up
GUNICORN_ENV_VAR_NAME = 'SERVER_SOFTWARE'


def __valid_platform():
    plat = sys.platform
    return plat in PLATFORMS_WITH_FORK_SUPPORT


def __valid_implementation():
    implementation = platform.python_implementation()
    return implementation == PYTHON_IMPLEMENTATION_WITH_FORK_SUPPORT


def __is_gunicorn():
    return 'gunicorn' in environ.get(GUNICORN_ENV_VAR_NAME, '')


def __is_gunicorn_master_proc():
    if not __is_gunicorn():
        return False
    full_stack = inspect.stack(0)
    for frame in full_stack:
        # Starting a worker
        if 'gunicorn/workers/base.py' in frame[1]:
            return False
    return True


def __is_billiard_master_proc():
    try:
        billiard_process = sys.modules['billiard.process']
        return billiard_process.current_process().name == 'MainProcess'
    except KeyError:
        return False


def detect_multiprocessing():
    if not __valid_platform() or not __valid_implementation():
        return False
    is_gunicorn_master = __is_gunicorn_master_proc()
    # Billiard includes Celery
    is_billiard_master = __is_billiard_master_proc()
    return is_gunicorn_master or is_billiard_master



from .namespace import Namespace
from .python_object_namespace import PythonObjectNamespace


class NoopNamespace(Namespace):
    def __init__(self):
        super(NoopNamespace, self).__init__(self.METHODS)
        self.span = None

    def call_method(self, name, args):
        return PythonObjectNamespace(None)

    def read_key(self, key):
        return PythonObjectNamespace(None)

    def read_attribute(self, name):
        return PythonObjectNamespace(None)

    METHODS = ()



# -*- coding: utf-8 -*-
import os
import types

import six
from functools import partial

from .exts.cloud_debug_python.module_explorer import GetCodeObjectAtLine, _GetModuleCodeObjects, _GetLineNumbers

from rook.logger import logger
from rook.singleton import singleton_obj
from rook.processor.error import Error

from ..exceptions import RookBdbCodeNotFound, RookBdbSetBreakpointFailed, RookInvalidPositionException, \
    RookSettingBreakpointError

try:
    from . import cdbg_native
    cdbg_native.InitializeModule(None)
except ImportError:
    # Special handling for Google AppEngine (Python 2.7)
    from google.devtools.cdbg.debuglets.python import cdbg_native


class BPStatus(object):
    __slots__ = ["disabled"]

    def __init__(self):
        self.disabled = False

class Bdb(object):
    def __init__(self):
        self.fncache = {}
        self._cookies = {}
        self._bp_status = {}
        self.user_line = None

    def set_trace(self):
        # Not needed
        pass

    def canonic(self, filename):
        if filename[0] == "<" and filename[-1] == ">":
            return filename
        canonic = self.fncache.get(filename)
        if not canonic:
            canonic = os.path.abspath(filename)
            canonic = os.path.normpath(canonic)
            self.fncache[filename] = canonic
        return canonic

    def ignore_current_thread(self):
        # Not needed
        pass

    def set_break(self, item, filename, lineno, aug_id):
        filename = self.canonic(filename)

        if isinstance(item, types.ModuleType):
            self._set_break_module(item, filename, lineno, aug_id)
        elif isinstance(item, types.CodeType):
            # the caller doesn't know if the code object has this line, so verify here
            self._set_break_code_object(item, filename, lineno, aug_id)
        else:
            raise KeyError(type(item))

    def _set_break_module(self, module, filename, lineno, aug_id):
        status, code_object = GetCodeObjectAtLine(module, lineno)
        if not status:
            if hasattr(module, '__file__'):
                logger.debug("CodeNotFound module filename %s", module.__file__)

            for cobj in _GetModuleCodeObjects(module):
                logger.debug("Name: %s", cobj.co_name)
                for cline in _GetLineNumbers(cobj):
                    logger.debug("Name: %s, Line %d", cobj.co_name, cline)

            if code_object == (None, None):
                raise RookBdbCodeNotFound(filename=filename)
            else:
                raise RookInvalidPositionException(filename=filename, line=lineno, alternatives=code_object)

        self._set_break_code_object(code_object, filename, lineno, aug_id)

    def _set_break_code_object(self, code_object, filename, lineno, aug_id):
        # Install the breakpoint
        bp_status = BPStatus()
        callback_partial = self._get_callback_partial(callback, lineno, code_object, bp_status, filename, aug_id)
        error_callback_partial = self._get_callback_partial(error_callback, lineno, code_object, bp_status, filename, aug_id)

        cookie = cdbg_native.SetConditionalBreakpoint(code_object, lineno, None, callback_partial,
                                                      error_callback_partial)

        if cookie < 0:
            raise RookBdbSetBreakpointFailed("%s on line %d" % (code_object.co_name, lineno))

        self._cookies[aug_id] = cookie
        self._bp_status[aug_id] = bp_status

    def _get_callback_partial(self, callback, lineno, code_object, bp_status, filename, aug_id):
        return partial(callback, lineno=lineno,
                user_line=self.user_line,
                callback_object_id=id(
                    code_object),
                bp_status=bp_status,
                filename=filename,
                pid=os.getpid(),
                aug_id=aug_id)

    def clear_break(self, aug_id):
        try:
            cookie = self._cookies[aug_id]
            status = self._bp_status[aug_id]
        except KeyError:
            return

        cdbg_native.ClearConditionalBreakpoint(cookie)
        status.disabled = True
        del self._cookies[aug_id]
        del self._bp_status[aug_id]

    def clear_all_breaks(self):
        for cookie in six.itervalues(self._cookies):
            cdbg_native.ClearConditionalBreakpoint(cookie)

        for status in six.itervalues(self._bp_status):
            status.disabled = True

        self._cookies = {}
        self._bp_status = {}

    def close(self):
        pass


def error_callback(lineno, user_line, callback_object_id, bp_status, filename, pid, aug_id):
    message = "Line: {lineno}, not found in file: {filename}, please contact support.".format(
        lineno=lineno, filename=filename)
    send_error_for_aug(aug_id, message)


def send_error_for_aug(aug_id, message):
    logger.error(message)
    if hasattr(singleton_obj, '_output'):
        exception = RookSettingBreakpointError(message)
        error = Error(exc=exception, message=message)
        singleton_obj._output.send_rule_status(aug_id, 'Error', error)


# This function has been moved outside of the class so that it can be pickled
# safely by cloudpickle (which will pickle any objects referred to by its closure)
# When changing it, take care to avoid using references to anything not imported within the function
def callback(lineno, user_line, callback_object_id, bp_status, filename, pid, aug_id):
    try:
        if bp_status.disabled is True:
            return

        import inspect
        frame = inspect.currentframe().f_back

        callback_was_pickled = callback_object_id != id(frame.f_code) or pid != os.getpid()

        if not callback_was_pickled and frame and user_line:
            user_line(frame, filename, lineno=lineno, aug_id=aug_id)
    except:  # this is very last line of defense in the hook, can't report from here lgtm[py/catch-base-exception]
        pass



from rook.logger import logger
from rook.user_warnings import UserWarnings
from rook.processor.error import Error


class Location(object):
    def __init__(self, output, aug):
        self._output = output
        self._aug = aug

        self._status = None
        self._warningCache = {}
        self._logCache = {}

    @property
    def aug_id(self):
        return self._aug.aug_id

    def add_aug(self, trigger_services):
        """Use the location to add the Aug to the relevant trigger service."""
        try:
            self._add_aug_internal(trigger_services)
        except Exception as exc:
            message = "Exception when adding aug"
            logger.exception(message + " " + self.aug_id + " : " + str(exc))
            self.set_error(Error(exc=exc, message=message))

    def _add_aug_internal(self, trigger_services):
        raise NotImplementedError()

    def execute(self, frame, extracted):
        with UserWarnings(self):
            try:
                self._aug.execute(frame, extracted, self._output)
            # Don't stop test exceptions from propagating
            except AssertionError:
                raise
            # Catch and silence everything else
            except Exception as exc:
                message = "Exception while processing Aug"
                rook_error = Error(exc=exc, message=message)

                if not self._should_silence_log(rook_error, self._logCache):
                    logger.exception(message)

                self.send_warning(rook_error)

    def set_active(self):
        self._send_rule_status("Active")

    def set_pending(self):
        self._send_rule_status("Pending")

    def set_removed(self):
        self._send_rule_status("Deleted")

    def set_error(self, error):
        self._send_rule_status("Error", error)

    def set_unknown(self, error):
        self._send_rule_status("Unknown", error)

    def send_warning(self, error):
        if self._should_silence_log(error, self._warningCache):
            return

        logger.warning(error.message)

        self._output.send_warning(self.aug_id, error)

    def _send_rule_status(self, status, error=None):
        if self._status == status:
            return

        logger.info("Updating rule status for %s to %s", self.aug_id, status)

        self._status = status
        self._output.send_rule_status(self.aug_id, status, error)

    def _should_silence_log(self, error, log_cache):
        if error.message in log_cache or len(log_cache) >= 10:
            return True

        log_cache[error.message] = True

        return False


import six
from rook.com_ws.envelope_wrappers.i_envelope_wrapper import IEnvelopeWrapper
from rook.processor.namespace_serializer2 import NamespaceSerializer2
from rook.protobuf.messages_pb2 import AugReportMessage


class Protobuf2EnvelopeWrapper(IEnvelopeWrapper):
    def __init__(self, agent_id, aug_id, message_id, arguments):
        self.message = AugReportMessage()
        self.message.agent_id = agent_id
        self.message.aug_id = aug_id
        self.message.report_id = message_id

        self.serializer = NamespaceSerializer2()
        self.serializer.dump(arguments, self.message.arguments2)

        self.buffer = None
        self.estimated_length = self.serializer.get_estimated_pending_bytes()

    def get_buffer(self):
        if self.buffer is not None:
            return self.buffer

        for k, v in six.iteritems(self.serializer.get_buffer_cache()):
            self.message.buffer_cache_indexes.append(v)
            self.message.buffer_cache_buffers.append(k)

        for k, v in six.iteritems(self.serializer.get_string_cache()):
            self.message.strings_cache[k] = v

        self.buffer = IEnvelopeWrapper.get_serialized_envelope(self.message)

        self.serializer = None
        self.message = None

        return self.buffer

    def __len__(self):      
        return self.estimated_length



import socket

import rook.com_ws.socketpair_compat  # do not remove - adds socket.socketpair on Windows lgtm[py/unused-import]
from rook.com_ws import poll_select
from threading import Event


class SelectableEvent(object):
    """
    Uses a socketpair to signal when the event is ready. This way, it can be used in a select.
    """
    def __init__(self):
        self._event = Event()
        self._readsocket, self._writesocket = socket.socketpair()
        self._readwaiter = poll_select.Waiter([self._readsocket], [], [])

    def close(self):
        self._readsocket.close()
        self._writesocket.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def fileno(self):
        return self._readsocket.fileno()

    def is_set(self):
        return self._event.is_set()

    def set(self):
        if not self.is_set():
            self._event.set()
            self._writesocket.send(b'1')

    def wait(self, timeout=None):
        if self._event.is_set():
            return True
        rfds, _, _ = self._readwaiter.wait(timeout)
        if len(rfds) != 0 and self._event.is_set():
            return True

        return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import variant_pb2 as variant__pb2
from . import variant2_pb2 as variant2__pb2
from . import agent_info_pb2 as agent__info__pb2
from . import controller_info_pb2 as controller__info__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0emessages.proto\x12\x0b\x63om.rookout\x1a\rvariant.proto\x1a\x0evariant2.proto\x1a\x10\x61gent_info.proto\x1a\x15\x63ontroller_info.proto\x1a\x19google/protobuf/any.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"D\n\x0fNewAgentMessage\x12\x31\n\nagent_info\x18\x01 \x01(\x0b\x32\x1d.com.rookout.AgentInformation\">\n\nStackFrame\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\x12\x0c\n\x04line\x18\x02 \x01(\r\x12\x10\n\x08\x66unction\x18\x03 \x01(\t\"5\n\nStackTrace\x12\'\n\x06\x66rames\x18\x01 \x03(\x0b\x32\x17.com.rookout.StackFrame\"~\n\tException\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12&\n\x08instance\x18\x03 \x01(\x0b\x32\x14.com.rookout.Variant\x12*\n\ttraceback\x18\x04 \x01(\x0b\x32\x17.com.rookout.StackTrace\"\xe2\x03\n\nLogMessage\x12-\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08\x61gent_id\x18\x02 \x01(\t\x12/\n\x05level\x18\x03 \x01(\x0e\x32 .com.rookout.LogMessage.LogLevel\x12\x10\n\x08\x66ilename\x18\x04 \x01(\t\x12\x0c\n\x04line\x18\x05 \x01(\r\x12\x0c\n\x04text\x18\x06 \x01(\t\x12\'\n\targuments\x18\x07 \x03(\x0b\x32\x14.com.rookout.Variant\x12)\n\texception\x18\x08 \x01(\x0b\x32\x16.com.rookout.Exception\x12\x1d\n\x11\x66ormatted_message\x18\t \x01(\tB\x02\x18\x01\x12\x32\n\x10legacy_arguments\x18\n \x01(\x0b\x32\x14.com.rookout.VariantB\x02\x18\x01\x12\x12\n\nclass_name\x18\x0b \x01(\t\x12\x13\n\x0bmethod_name\x18\x0c \x01(\t\x12\x15\n\rcontroller_id\x18\r \x01(\t\"M\n\x08LogLevel\x12\t\n\x05TRACE\x10\x00\x12\t\n\x05\x44\x45\x42UG\x10\x01\x12\x08\n\x04INFO\x10\x02\x12\x0b\n\x07WARNING\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\t\n\x05\x46\x41TAL\x10\x05\"\xa3\x05\n\x14\x43ontrollerLogMessage\x12-\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x39\n\x05level\x18\x02 \x01(\x0e\x32*.com.rookout.ControllerLogMessage.LogLevel\x12\x10\n\x08\x66ilename\x18\x03 \x01(\t\x12\x0c\n\x04line\x18\x04 \x01(\r\x12\x0c\n\x04text\x18\x05 \x01(\t\x12\x12\n\nclass_name\x18\x06 \x01(\t\x12\x13\n\x0bmethod_name\x18\x07 \x01(\t\x12\x15\n\rcontroller_id\x18\x08 \x01(\t\x12\x36\n\x05\x65rror\x18\t \x01(\x0b\x32\'.com.rookout.ControllerLogMessage.Error\x12\x41\n\x08metadata\x18\n \x03(\x0b\x32/.com.rookout.ControllerLogMessage.MetadataEntry\x12U\n\x12metadataJSONValues\x18\x0b \x03(\x0b\x32\x39.com.rookout.ControllerLogMessage.MetadataJSONValuesEntry\x1a&\n\x05\x45rror\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x39\n\x17MetadataJSONValuesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\"M\n\x08LogLevel\x12\t\n\x05TRACE\x10\x00\x12\t\n\x05\x44\x45\x42UG\x10\x01\x12\x08\n\x04INFO\x10\x02\x12\x0b\n\x07WARNING\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\t\n\x05\x46\x41TAL\x10\x05\"\x80\x01\n\x11RuleStatusMessage\x12\x10\n\x08\x61gent_id\x18\x01 \x01(\t\x12\x0f\n\x07rule_id\x18\x02 \x01(\t\x12\x0e\n\x06\x61\x63tive\x18\x03 \x01(\t\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.com.rookout.Error\x12\x15\n\rcontroller_id\x18\x05 \x01(\t\"\xf0\x02\n\x10\x41ugReportMessage\x12\x10\n\x08\x61gent_id\x18\x01 \x01(\t\x12\x0e\n\x06\x61ug_id\x18\x02 \x01(\t\x12\'\n\targuments\x18\x03 \x01(\x0b\x32\x14.com.rookout.Variant\x12\x11\n\treport_id\x18\x04 \x01(\t\x12\x46\n\rstrings_cache\x18\x05 \x03(\x0b\x32/.com.rookout.AugReportMessage.StringsCacheEntry\x12\x1c\n\x14\x62uffer_cache_buffers\x18\x06 \x03(\x0c\x12\x1c\n\x14\x62uffer_cache_indexes\x18\x07 \x03(\r\x12)\n\narguments2\x18\x08 \x01(\x0b\x32\x15.com.rookout.Variant2\x12\x1a\n\x12reverse_list_order\x18\t \x01(\x08\x1a\x33\n\x11StringsCacheEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\"e\n\x07UserMsg\x12\x31\n\naug_report\x18\x01 \x01(\x0b\x32\x1d.com.rookout.AugReportMessage\x12\x11\n\treport_id\x18\x02 \x01(\t\x12\x14\n\x0cworkspace_id\x18\x03 \x01(\t\"!\n\rAddAugCommand\x12\x10\n\x08\x61ug_json\x18\x01 \x01(\t\"\xad\x01\n\x12InitialAugsCommand\x12\x0c\n\x04\x61ugs\x18\x01 \x03(\t\x12P\n\x11sdk_configuration\x18\x02 \x03(\x0b\x32\x35.com.rookout.InitialAugsCommand.SdkConfigurationEntry\x1a\x37\n\x15SdkConfigurationEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\"\n\x10RemoveAugCommand\x12\x0e\n\x06\x61ug_id\x18\x01 \x01(\t\"\x15\n\x13InitFinishedMessage\"\x12\n\x10\x43learAugsCommand\"\x1c\n\x0bPingMessage\x12\r\n\x05value\x18\x01 \x01(\x05\"\x1f\n\x0c\x45rrorMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\"e\n\x0e\x43\x32\x43InitMessage\x12:\n\x0e\x63ontrollerInfo\x18\x01 \x01(\x0b\x32\".com.rookout.ControllerInformation\x12\x17\n\x0f\x63onnectedAgents\x18\x02 \x03(\t\"U\n\x16\x44\x61taOnPremTokenRequest\x12;\n\x0b\x61\x63\x63\x65ss_type\x18\x01 \x01(\x0e\x32&.com.rookout.DataOnPremTokenAccessType\"(\n\x17\x44\x61taOnPremTokenResponse\x12\r\n\x05token\x18\x01 \x01(\x0c\" \n\nAgentsList\x12\x12\n\nagents_ids\x18\x01 \x03(\t\"!\n\x0c\x41ssertAgents\x12\x11\n\tagent_ids\x18\x01 \x03(\t\"-\n\x17InvalidateAgentsCommand\x12\x12\n\nagents_ids\x18\x01 \x03(\t\"*\n\x14\x43onfirmAgentsCommand\x12\x12\n\nagents_ids\x18\x01 \x03(\t\":\n\x13\x41gentIdHitCountPair\x12\x10\n\x08\x61gent_id\x18\x01 \x01(\t\x12\x11\n\thit_count\x18\x02 \x01(\r\"f\n\x15HitCountUpdateMessage\x12\x0e\n\x06\x61ug_id\x18\x01 \x01(\t\x12=\n\x13hitCountInformation\x18\x02 \x03(\x0b\x32 .com.rookout.AgentIdHitCountPair\"L\n\rC2cRpcRequest\x12\x14\n\x0crequest_uuid\x18\x01 \x01(\t\x12%\n\x07message\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\"\\\n\x0e\x43\x32\x63RpcResponse\x12\x14\n\x0crequest_uuid\x18\x01 \x01(\t\x12%\n\x07message\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x12\r\n\x05\x65rror\x18\x03 \x01(\t\"w\n%StartDatastoreConnectivityTestCommand\x12\r\n\x05orgId\x18\x01 \x01(\t\x12\x14\n\x0c\x63ontrollerId\x18\x02 \x01(\t\x12\x14\n\x0c\x64\x61tastoreUrl\x18\x03 \x01(\t\x12\x13\n\x0bworkspaceId\x18\x04 \x01(\t\"{\n$DatastoreConnectivityTestC2cResponse\x12\x14\n\x0c\x63ontrollerId\x18\x01 \x01(\t\x12\x15\n\rTestMessageId\x18\x02 \x01(\t\x12\x11\n\tisSuccess\x18\x03 \x01(\x08\x12\x13\n\x0b\x65rrorString\x18\x04 \x01(\t\",\n\x14PingDatastoreCommand\x12\x14\n\x0c\x64\x61tastoreUrl\x18\x01 \x01(\t\"\xb3\x01\n\x14\x44\x61tastoreInformation\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x10\n\x08hostname\x18\x02 \x01(\t\x12\x14\n\x0cmachine_type\x18\x03 \x01(\t\x12\x0f\n\x07network\x18\x04 \x01(\t\x12\n\n\x02ip\x18\x05 \x01(\t\x12\n\n\x02os\x18\x06 \x01(\t\x12\x12\n\nos_release\x18\x07 \x01(\t\x12\x12\n\nos_version\x18\x08 \x01(\t\x12\x11\n\tos_string\x18\t \x01(\t*0\n\x19\x44\x61taOnPremTokenAccessType\x12\t\n\x05WRITE\x10\x00\x12\x08\n\x04READ\x10\x01\x42O\n\x19\x63om.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2b\x06proto3')

_DATAONPREMTOKENACCESSTYPE = DESCRIPTOR.enum_types_by_name['DataOnPremTokenAccessType']
DataOnPremTokenAccessType = enum_type_wrapper.EnumTypeWrapper(_DATAONPREMTOKENACCESSTYPE)
WRITE = 0
READ = 1


_NEWAGENTMESSAGE = DESCRIPTOR.message_types_by_name['NewAgentMessage']
_STACKFRAME = DESCRIPTOR.message_types_by_name['StackFrame']
_STACKTRACE = DESCRIPTOR.message_types_by_name['StackTrace']
_EXCEPTION = DESCRIPTOR.message_types_by_name['Exception']
_LOGMESSAGE = DESCRIPTOR.message_types_by_name['LogMessage']
_CONTROLLERLOGMESSAGE = DESCRIPTOR.message_types_by_name['ControllerLogMessage']
_CONTROLLERLOGMESSAGE_ERROR = _CONTROLLERLOGMESSAGE.nested_types_by_name['Error']
_CONTROLLERLOGMESSAGE_METADATAENTRY = _CONTROLLERLOGMESSAGE.nested_types_by_name['MetadataEntry']
_CONTROLLERLOGMESSAGE_METADATAJSONVALUESENTRY = _CONTROLLERLOGMESSAGE.nested_types_by_name['MetadataJSONValuesEntry']
_RULESTATUSMESSAGE = DESCRIPTOR.message_types_by_name['RuleStatusMessage']
_AUGREPORTMESSAGE = DESCRIPTOR.message_types_by_name['AugReportMessage']
_AUGREPORTMESSAGE_STRINGSCACHEENTRY = _AUGREPORTMESSAGE.nested_types_by_name['StringsCacheEntry']
_USERMSG = DESCRIPTOR.message_types_by_name['UserMsg']
_ADDAUGCOMMAND = DESCRIPTOR.message_types_by_name['AddAugCommand']
_INITIALAUGSCOMMAND = DESCRIPTOR.message_types_by_name['InitialAugsCommand']
_INITIALAUGSCOMMAND_SDKCONFIGURATIONENTRY = _INITIALAUGSCOMMAND.nested_types_by_name['SdkConfigurationEntry']
_REMOVEAUGCOMMAND = DESCRIPTOR.message_types_by_name['RemoveAugCommand']
_INITFINISHEDMESSAGE = DESCRIPTOR.message_types_by_name['InitFinishedMessage']
_CLEARAUGSCOMMAND = DESCRIPTOR.message_types_by_name['ClearAugsCommand']
_PINGMESSAGE = DESCRIPTOR.message_types_by_name['PingMessage']
_ERRORMESSAGE = DESCRIPTOR.message_types_by_name['ErrorMessage']
_C2CINITMESSAGE = DESCRIPTOR.message_types_by_name['C2CInitMessage']
_DATAONPREMTOKENREQUEST = DESCRIPTOR.message_types_by_name['DataOnPremTokenRequest']
_DATAONPREMTOKENRESPONSE = DESCRIPTOR.message_types_by_name['DataOnPremTokenResponse']
_AGENTSLIST = DESCRIPTOR.message_types_by_name['AgentsList']
_ASSERTAGENTS = DESCRIPTOR.message_types_by_name['AssertAgents']
_INVALIDATEAGENTSCOMMAND = DESCRIPTOR.message_types_by_name['InvalidateAgentsCommand']
_CONFIRMAGENTSCOMMAND = DESCRIPTOR.message_types_by_name['ConfirmAgentsCommand']
_AGENTIDHITCOUNTPAIR = DESCRIPTOR.message_types_by_name['AgentIdHitCountPair']
_HITCOUNTUPDATEMESSAGE = DESCRIPTOR.message_types_by_name['HitCountUpdateMessage']
_C2CRPCREQUEST = DESCRIPTOR.message_types_by_name['C2cRpcRequest']
_C2CRPCRESPONSE = DESCRIPTOR.message_types_by_name['C2cRpcResponse']
_STARTDATASTORECONNECTIVITYTESTCOMMAND = DESCRIPTOR.message_types_by_name['StartDatastoreConnectivityTestCommand']
_DATASTORECONNECTIVITYTESTC2CRESPONSE = DESCRIPTOR.message_types_by_name['DatastoreConnectivityTestC2cResponse']
_PINGDATASTORECOMMAND = DESCRIPTOR.message_types_by_name['PingDatastoreCommand']
_DATASTOREINFORMATION = DESCRIPTOR.message_types_by_name['DatastoreInformation']
_LOGMESSAGE_LOGLEVEL = _LOGMESSAGE.enum_types_by_name['LogLevel']
_CONTROLLERLOGMESSAGE_LOGLEVEL = _CONTROLLERLOGMESSAGE.enum_types_by_name['LogLevel']
NewAgentMessage = _reflection.GeneratedProtocolMessageType('NewAgentMessage', (_message.Message,), {
  'DESCRIPTOR' : _NEWAGENTMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.NewAgentMessage)
  })
_sym_db.RegisterMessage(NewAgentMessage)

StackFrame = _reflection.GeneratedProtocolMessageType('StackFrame', (_message.Message,), {
  'DESCRIPTOR' : _STACKFRAME,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.StackFrame)
  })
_sym_db.RegisterMessage(StackFrame)

StackTrace = _reflection.GeneratedProtocolMessageType('StackTrace', (_message.Message,), {
  'DESCRIPTOR' : _STACKTRACE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.StackTrace)
  })
_sym_db.RegisterMessage(StackTrace)

Exception = _reflection.GeneratedProtocolMessageType('Exception', (_message.Message,), {
  'DESCRIPTOR' : _EXCEPTION,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.Exception)
  })
_sym_db.RegisterMessage(Exception)

LogMessage = _reflection.GeneratedProtocolMessageType('LogMessage', (_message.Message,), {
  'DESCRIPTOR' : _LOGMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.LogMessage)
  })
_sym_db.RegisterMessage(LogMessage)

ControllerLogMessage = _reflection.GeneratedProtocolMessageType('ControllerLogMessage', (_message.Message,), {

  'Error' : _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), {
    'DESCRIPTOR' : _CONTROLLERLOGMESSAGE_ERROR,
    '__module__' : 'messages_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.ControllerLogMessage.Error)
    })
  ,

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _CONTROLLERLOGMESSAGE_METADATAENTRY,
    '__module__' : 'messages_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.ControllerLogMessage.MetadataEntry)
    })
  ,

  'MetadataJSONValuesEntry' : _reflection.GeneratedProtocolMessageType('MetadataJSONValuesEntry', (_message.Message,), {
    'DESCRIPTOR' : _CONTROLLERLOGMESSAGE_METADATAJSONVALUESENTRY,
    '__module__' : 'messages_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.ControllerLogMessage.MetadataJSONValuesEntry)
    })
  ,
  'DESCRIPTOR' : _CONTROLLERLOGMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.ControllerLogMessage)
  })
_sym_db.RegisterMessage(ControllerLogMessage)
_sym_db.RegisterMessage(ControllerLogMessage.Error)
_sym_db.RegisterMessage(ControllerLogMessage.MetadataEntry)
_sym_db.RegisterMessage(ControllerLogMessage.MetadataJSONValuesEntry)

RuleStatusMessage = _reflection.GeneratedProtocolMessageType('RuleStatusMessage', (_message.Message,), {
  'DESCRIPTOR' : _RULESTATUSMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.RuleStatusMessage)
  })
_sym_db.RegisterMessage(RuleStatusMessage)

AugReportMessage = _reflection.GeneratedProtocolMessageType('AugReportMessage', (_message.Message,), {

  'StringsCacheEntry' : _reflection.GeneratedProtocolMessageType('StringsCacheEntry', (_message.Message,), {
    'DESCRIPTOR' : _AUGREPORTMESSAGE_STRINGSCACHEENTRY,
    '__module__' : 'messages_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.AugReportMessage.StringsCacheEntry)
    })
  ,
  'DESCRIPTOR' : _AUGREPORTMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AugReportMessage)
  })
_sym_db.RegisterMessage(AugReportMessage)
_sym_db.RegisterMessage(AugReportMessage.StringsCacheEntry)

UserMsg = _reflection.GeneratedProtocolMessageType('UserMsg', (_message.Message,), {
  'DESCRIPTOR' : _USERMSG,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.UserMsg)
  })
_sym_db.RegisterMessage(UserMsg)

AddAugCommand = _reflection.GeneratedProtocolMessageType('AddAugCommand', (_message.Message,), {
  'DESCRIPTOR' : _ADDAUGCOMMAND,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AddAugCommand)
  })
_sym_db.RegisterMessage(AddAugCommand)

InitialAugsCommand = _reflection.GeneratedProtocolMessageType('InitialAugsCommand', (_message.Message,), {

  'SdkConfigurationEntry' : _reflection.GeneratedProtocolMessageType('SdkConfigurationEntry', (_message.Message,), {
    'DESCRIPTOR' : _INITIALAUGSCOMMAND_SDKCONFIGURATIONENTRY,
    '__module__' : 'messages_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.InitialAugsCommand.SdkConfigurationEntry)
    })
  ,
  'DESCRIPTOR' : _INITIALAUGSCOMMAND,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.InitialAugsCommand)
  })
_sym_db.RegisterMessage(InitialAugsCommand)
_sym_db.RegisterMessage(InitialAugsCommand.SdkConfigurationEntry)

RemoveAugCommand = _reflection.GeneratedProtocolMessageType('RemoveAugCommand', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEAUGCOMMAND,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.RemoveAugCommand)
  })
_sym_db.RegisterMessage(RemoveAugCommand)

InitFinishedMessage = _reflection.GeneratedProtocolMessageType('InitFinishedMessage', (_message.Message,), {
  'DESCRIPTOR' : _INITFINISHEDMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.InitFinishedMessage)
  })
_sym_db.RegisterMessage(InitFinishedMessage)

ClearAugsCommand = _reflection.GeneratedProtocolMessageType('ClearAugsCommand', (_message.Message,), {
  'DESCRIPTOR' : _CLEARAUGSCOMMAND,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.ClearAugsCommand)
  })
_sym_db.RegisterMessage(ClearAugsCommand)

PingMessage = _reflection.GeneratedProtocolMessageType('PingMessage', (_message.Message,), {
  'DESCRIPTOR' : _PINGMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.PingMessage)
  })
_sym_db.RegisterMessage(PingMessage)

ErrorMessage = _reflection.GeneratedProtocolMessageType('ErrorMessage', (_message.Message,), {
  'DESCRIPTOR' : _ERRORMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.ErrorMessage)
  })
_sym_db.RegisterMessage(ErrorMessage)

C2CInitMessage = _reflection.GeneratedProtocolMessageType('C2CInitMessage', (_message.Message,), {
  'DESCRIPTOR' : _C2CINITMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.C2CInitMessage)
  })
_sym_db.RegisterMessage(C2CInitMessage)

DataOnPremTokenRequest = _reflection.GeneratedProtocolMessageType('DataOnPremTokenRequest', (_message.Message,), {
  'DESCRIPTOR' : _DATAONPREMTOKENREQUEST,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.DataOnPremTokenRequest)
  })
_sym_db.RegisterMessage(DataOnPremTokenRequest)

DataOnPremTokenResponse = _reflection.GeneratedProtocolMessageType('DataOnPremTokenResponse', (_message.Message,), {
  'DESCRIPTOR' : _DATAONPREMTOKENRESPONSE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.DataOnPremTokenResponse)
  })
_sym_db.RegisterMessage(DataOnPremTokenResponse)

AgentsList = _reflection.GeneratedProtocolMessageType('AgentsList', (_message.Message,), {
  'DESCRIPTOR' : _AGENTSLIST,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AgentsList)
  })
_sym_db.RegisterMessage(AgentsList)

AssertAgents = _reflection.GeneratedProtocolMessageType('AssertAgents', (_message.Message,), {
  'DESCRIPTOR' : _ASSERTAGENTS,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AssertAgents)
  })
_sym_db.RegisterMessage(AssertAgents)

InvalidateAgentsCommand = _reflection.GeneratedProtocolMessageType('InvalidateAgentsCommand', (_message.Message,), {
  'DESCRIPTOR' : _INVALIDATEAGENTSCOMMAND,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.InvalidateAgentsCommand)
  })
_sym_db.RegisterMessage(InvalidateAgentsCommand)

ConfirmAgentsCommand = _reflection.GeneratedProtocolMessageType('ConfirmAgentsCommand', (_message.Message,), {
  'DESCRIPTOR' : _CONFIRMAGENTSCOMMAND,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.ConfirmAgentsCommand)
  })
_sym_db.RegisterMessage(ConfirmAgentsCommand)

AgentIdHitCountPair = _reflection.GeneratedProtocolMessageType('AgentIdHitCountPair', (_message.Message,), {
  'DESCRIPTOR' : _AGENTIDHITCOUNTPAIR,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AgentIdHitCountPair)
  })
_sym_db.RegisterMessage(AgentIdHitCountPair)

HitCountUpdateMessage = _reflection.GeneratedProtocolMessageType('HitCountUpdateMessage', (_message.Message,), {
  'DESCRIPTOR' : _HITCOUNTUPDATEMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.HitCountUpdateMessage)
  })
_sym_db.RegisterMessage(HitCountUpdateMessage)

C2cRpcRequest = _reflection.GeneratedProtocolMessageType('C2cRpcRequest', (_message.Message,), {
  'DESCRIPTOR' : _C2CRPCREQUEST,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.C2cRpcRequest)
  })
_sym_db.RegisterMessage(C2cRpcRequest)

C2cRpcResponse = _reflection.GeneratedProtocolMessageType('C2cRpcResponse', (_message.Message,), {
  'DESCRIPTOR' : _C2CRPCRESPONSE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.C2cRpcResponse)
  })
_sym_db.RegisterMessage(C2cRpcResponse)

StartDatastoreConnectivityTestCommand = _reflection.GeneratedProtocolMessageType('StartDatastoreConnectivityTestCommand', (_message.Message,), {
  'DESCRIPTOR' : _STARTDATASTORECONNECTIVITYTESTCOMMAND,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.StartDatastoreConnectivityTestCommand)
  })
_sym_db.RegisterMessage(StartDatastoreConnectivityTestCommand)

DatastoreConnectivityTestC2cResponse = _reflection.GeneratedProtocolMessageType('DatastoreConnectivityTestC2cResponse', (_message.Message,), {
  'DESCRIPTOR' : _DATASTORECONNECTIVITYTESTC2CRESPONSE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.DatastoreConnectivityTestC2cResponse)
  })
_sym_db.RegisterMessage(DatastoreConnectivityTestC2cResponse)

PingDatastoreCommand = _reflection.GeneratedProtocolMessageType('PingDatastoreCommand', (_message.Message,), {
  'DESCRIPTOR' : _PINGDATASTORECOMMAND,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.PingDatastoreCommand)
  })
_sym_db.RegisterMessage(PingDatastoreCommand)

DatastoreInformation = _reflection.GeneratedProtocolMessageType('DatastoreInformation', (_message.Message,), {
  'DESCRIPTOR' : _DATASTOREINFORMATION,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.DatastoreInformation)
  })
_sym_db.RegisterMessage(DatastoreInformation)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2'
  _LOGMESSAGE.fields_by_name['formatted_message']._options = None
  _LOGMESSAGE.fields_by_name['formatted_message']._serialized_options = b'\030\001'
  _LOGMESSAGE.fields_by_name['legacy_arguments']._options = None
  _LOGMESSAGE.fields_by_name['legacy_arguments']._serialized_options = b'\030\001'
  _CONTROLLERLOGMESSAGE_METADATAENTRY._options = None
  _CONTROLLERLOGMESSAGE_METADATAENTRY._serialized_options = b'8\001'
  _CONTROLLERLOGMESSAGE_METADATAJSONVALUESENTRY._options = None
  _CONTROLLERLOGMESSAGE_METADATAJSONVALUESENTRY._serialized_options = b'8\001'
  _AUGREPORTMESSAGE_STRINGSCACHEENTRY._options = None
  _AUGREPORTMESSAGE_STRINGSCACHEENTRY._serialized_options = b'8\001'
  _INITIALAUGSCOMMAND_SDKCONFIGURATIONENTRY._options = None
  _INITIALAUGSCOMMAND_SDKCONFIGURATIONENTRY._serialized_options = b'8\001'
  _DATAONPREMTOKENACCESSTYPE._serialized_start=3803
  _DATAONPREMTOKENACCESSTYPE._serialized_end=3851
  _NEWAGENTMESSAGE._serialized_start=163
  _NEWAGENTMESSAGE._serialized_end=231
  _STACKFRAME._serialized_start=233
  _STACKFRAME._serialized_end=295
  _STACKTRACE._serialized_start=297
  _STACKTRACE._serialized_end=350
  _EXCEPTION._serialized_start=352
  _EXCEPTION._serialized_end=478
  _LOGMESSAGE._serialized_start=481
  _LOGMESSAGE._serialized_end=963
  _LOGMESSAGE_LOGLEVEL._serialized_start=886
  _LOGMESSAGE_LOGLEVEL._serialized_end=963
  _CONTROLLERLOGMESSAGE._serialized_start=966
  _CONTROLLERLOGMESSAGE._serialized_end=1641
  _CONTROLLERLOGMESSAGE_ERROR._serialized_start=1416
  _CONTROLLERLOGMESSAGE_ERROR._serialized_end=1454
  _CONTROLLERLOGMESSAGE_METADATAENTRY._serialized_start=1456
  _CONTROLLERLOGMESSAGE_METADATAENTRY._serialized_end=1503
  _CONTROLLERLOGMESSAGE_METADATAJSONVALUESENTRY._serialized_start=1505
  _CONTROLLERLOGMESSAGE_METADATAJSONVALUESENTRY._serialized_end=1562
  _CONTROLLERLOGMESSAGE_LOGLEVEL._serialized_start=886
  _CONTROLLERLOGMESSAGE_LOGLEVEL._serialized_end=963
  _RULESTATUSMESSAGE._serialized_start=1644
  _RULESTATUSMESSAGE._serialized_end=1772
  _AUGREPORTMESSAGE._serialized_start=1775
  _AUGREPORTMESSAGE._serialized_end=2143
  _AUGREPORTMESSAGE_STRINGSCACHEENTRY._serialized_start=2092
  _AUGREPORTMESSAGE_STRINGSCACHEENTRY._serialized_end=2143
  _USERMSG._serialized_start=2145
  _USERMSG._serialized_end=2246
  _ADDAUGCOMMAND._serialized_start=2248
  _ADDAUGCOMMAND._serialized_end=2281
  _INITIALAUGSCOMMAND._serialized_start=2284
  _INITIALAUGSCOMMAND._serialized_end=2457
  _INITIALAUGSCOMMAND_SDKCONFIGURATIONENTRY._serialized_start=2402
  _INITIALAUGSCOMMAND_SDKCONFIGURATIONENTRY._serialized_end=2457
  _REMOVEAUGCOMMAND._serialized_start=2459
  _REMOVEAUGCOMMAND._serialized_end=2493
  _INITFINISHEDMESSAGE._serialized_start=2495
  _INITFINISHEDMESSAGE._serialized_end=2516
  _CLEARAUGSCOMMAND._serialized_start=2518
  _CLEARAUGSCOMMAND._serialized_end=2536
  _PINGMESSAGE._serialized_start=2538
  _PINGMESSAGE._serialized_end=2566
  _ERRORMESSAGE._serialized_start=2568
  _ERRORMESSAGE._serialized_end=2599
  _C2CINITMESSAGE._serialized_start=2601
  _C2CINITMESSAGE._serialized_end=2702
  _DATAONPREMTOKENREQUEST._serialized_start=2704
  _DATAONPREMTOKENREQUEST._serialized_end=2789
  _DATAONPREMTOKENRESPONSE._serialized_start=2791
  _DATAONPREMTOKENRESPONSE._serialized_end=2831
  _AGENTSLIST._serialized_start=2833
  _AGENTSLIST._serialized_end=2865
  _ASSERTAGENTS._serialized_start=2867
  _ASSERTAGENTS._serialized_end=2900
  _INVALIDATEAGENTSCOMMAND._serialized_start=2902
  _INVALIDATEAGENTSCOMMAND._serialized_end=2947
  _CONFIRMAGENTSCOMMAND._serialized_start=2949
  _CONFIRMAGENTSCOMMAND._serialized_end=2991
  _AGENTIDHITCOUNTPAIR._serialized_start=2993
  _AGENTIDHITCOUNTPAIR._serialized_end=3051
  _HITCOUNTUPDATEMESSAGE._serialized_start=3053
  _HITCOUNTUPDATEMESSAGE._serialized_end=3155
  _C2CRPCREQUEST._serialized_start=3157
  _C2CRPCREQUEST._serialized_end=3233
  _C2CRPCRESPONSE._serialized_start=3235
  _C2CRPCRESPONSE._serialized_end=3327
  _STARTDATASTORECONNECTIVITYTESTCOMMAND._serialized_start=3329
  _STARTDATASTORECONNECTIVITYTESTCOMMAND._serialized_end=3448
  _DATASTORECONNECTIVITYTESTC2CRESPONSE._serialized_start=3450
  _DATASTORECONNECTIVITYTESTC2CRESPONSE._serialized_end=3573
  _PINGDATASTORECOMMAND._serialized_start=3575
  _PINGDATASTORECOMMAND._serialized_end=3619
  _DATASTOREINFORMATION._serialized_start=3622
  _DATASTOREINFORMATION._serialized_end=3801
# @@protoc_insertion_point(module_scope)



from threading import Lock


class AugRateLimiter(object):
    # enable this flag in unittests only to expose some methods that are needed for tests
    ENABLE_SPECIAL_METHODS_FOR_TESTS = False

    def __init__(self, quota, window_size):
        self._quota = quota
        self._window_size = window_size
        self._lock = Lock()

        self._windows = {}

    def allow(self, now_ns):
        if self._quota is None:
            return '0'

        with self._lock:
            self._cleanup(now_ns)

            current_window_key = now_ns // self._window_size * self._window_size
            prev_window_key = current_window_key - self._window_size

            current_window_usage = self._windows.setdefault(current_window_key, 0)
            prev_window_usage = self._windows.get(prev_window_key)

            if prev_window_usage is None:
                if current_window_usage > self._quota:
                    return None
            else:
                prev_weight = 1 - (now_ns - current_window_key) / float(self._window_size)
                weighted_usage = (prev_window_usage * prev_weight) + current_window_usage

                if weighted_usage > self._quota:
                    return None

            return current_window_key

    def record(self, key, duration_ns):
        if self._quota is None:
            return

        with self._lock:
            total_duration = self._windows.get(key)  # windows might be cleared while aug is running (unlikely)

            if total_duration is not None:
                self._windows[key] += duration_ns

    def _cleanup(self, now):
        if len(self._windows) > 10:
            self._windows = {k: v for k, v in self._windows.items() if k > now - self._window_size * 5}

    def get_windows_for_tests_only(self):
        if not AugRateLimiter.ENABLE_SPECIAL_METHODS_FOR_TESTS:
            raise NotImplementedError('This method should be used only in unit tests!!!!')
        return self._windows



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: controller_info.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x63ontroller_info.proto\x12\x0b\x63om.rookout\"\xf6\x01\n\x15\x43ontrollerInformation\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06\x63ommit\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x0c\n\x04tags\x18\x04 \x03(\t\x12\x10\n\x08hostname\x18\x05 \x01(\t\x12\x14\n\x0cmachine_type\x18\x06 \x01(\t\x12\x0f\n\x07network\x18\x07 \x01(\t\x12\n\n\x02ip\x18\x08 \x01(\t\x12\x10\n\x08platform\x18\t \x01(\t\x12\x18\n\x10platform_release\x18\n \x01(\t\x12\x18\n\x10platform_version\x18\x0b \x01(\t\x12\x17\n\x0fplatform_string\x18\x0c \x01(\tBO\n\x19\x63om.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2b\x06proto3')



_CONTROLLERINFORMATION = DESCRIPTOR.message_types_by_name['ControllerInformation']
ControllerInformation = _reflection.GeneratedProtocolMessageType('ControllerInformation', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLLERINFORMATION,
  '__module__' : 'controller_info_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.ControllerInformation)
  })
_sym_db.RegisterMessage(ControllerInformation)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2'
  _CONTROLLERINFORMATION._serialized_start=39
  _CONTROLLERINFORMATION._serialized_end=285
# @@protoc_insertion_point(module_scope)



"""This module is in charge of building a Aug object from it's over the wire form.

Since most serialization format do not support polymorphism, we add this capability in this module.

This module finds and loads all dynamic classes and dynamically builds the relevant component based on it's id.
"""

import re

import six, os

from rook.factory import Factory
from .aug_rate_limiter import AugRateLimiter

from ..augs import actions, conditions, locations
from ..augs.aug import Aug


from rook.rookout_json import json
from rook import config

from rook.processor.processor_factory import ProcessorFactory

from rook.exceptions import ToolException, RookAugInvalidKey, RookObjectNameMissing, RookUnknownObject, \
    RookInvalidObjectConfiguration, RookUnsupportedLocation, RookInvalidRateLimitConfiguration
import rook.utils as utils


class AugFactory(Factory):
    """This is the factory for building Augs by their configuration."""

    def __init__(self, output):
        """Initialize the class."""
        super(AugFactory, self).__init__()

        self._output = output
        self._processor_factory = ProcessorFactory([], [])

        self._load_dynamic_classes()

        self._global_rate_limiter = None
        self.init_global_rate_limit()

    def init_global_rate_limit(self):
        rate_limit_env = os.environ.get('ROOKOUT_GLOBAL_RATE_LIMIT', None)

        if rate_limit_env:
            # parse_rate_limit_string Throws if not valid and prevent the rook from start
            rate_limits_ns = AugFactory.parse_rate_limit_string(rate_limit_env, 0, 0, True)
            self._global_rate_limiter = AugRateLimiter(rate_limits_ns[0], rate_limits_ns[1])
        else:
            self._global_rate_limiter = AugRateLimiter(None, None)  # Disabled

    @staticmethod
    def get_dict_value(configuration, value, default_value):
        val = configuration.get(value)
        return val if val is not None else default_value

    def get_aug(self, configuration):
        """Returns an Aug object based on the given configuration."""
        aug_id = None
        try:
            aug_id = configuration['id']
        except KeyError as exc:
            six.raise_from(RookAugInvalidKey('id', json.dumps(configuration)), exc)

        condition = None
        conditional = configuration.get('conditional')
        if conditional:
            condition = conditions.IfCondition(conditional)

        try:
            action_dict = configuration['action']
        except KeyError as exc:
            six.raise_from(RookAugInvalidKey('action', json.dumps(configuration)), exc)
        action = self._get_dynamic_class(action_dict)

        limits_spec = configuration.get('rateLimit', "")

        rate_limits_ns = AugFactory.parse_rate_limit_string(limits_spec, 200, 5000)

        max_aug_execution_time = AugFactory.get_dict_value(configuration, 'maxAugTime', config.InstrumentationConfig.MAX_AUG_TIME_MS)
        max_aug_execution_time = utils.milliseconds_to_int_nano_seconds(max_aug_execution_time)

        aug = Aug(aug_id=aug_id,
                  condition=condition,
                  action=action,
                  max_aug_execution_time_ns=max_aug_execution_time,
                  rate_limits_ns=rate_limits_ns,
                  global_rate_limit=self._global_rate_limiter)

        try:
            location_dict = configuration['location']
        except KeyError as exc:
            six.raise_from(RookAugInvalidKey('location', json.dumps(configuration)), exc)
        location = self._get_location(aug, location_dict)

        return location

    @staticmethod
    def parse_rate_limit_string(limits_spec, default_quota, default_window, throw=False):
        rate_limits = None

        r = re.match(r'^(\d+)[\\/](\d+)$', limits_spec)

        if r is not None:
            rate_limits = (int(r.group(1)), int(r.group(2)))

        if not rate_limits:
            rate_limits = (default_quota, default_window)

        if rate_limits[0] >= rate_limits[1] or r is None:
            if throw:
                raise RookInvalidRateLimitConfiguration(limits_spec)

        rate_limits_ns = tuple(utils.milliseconds_to_int_nano_seconds(r) for r in rate_limits)

        return rate_limits_ns

    def _load_dynamic_classes(self):
        """Load all dynamic classes into the factory."""
        self.register_methods(locations.__all__)
        self.register_methods(conditions.__all__)
        self.register_methods(actions.__all__)

    def _get_dynamic_class(self, configuration):
        """Return a class instance based on configuration."""

        if not configuration:
            return None
        else:
            factory = self._get_factory(configuration)

            try:
                return factory(configuration, self._processor_factory)
            except ToolException:
                raise
            except Exception as exc:
                six.raise_from(RookInvalidObjectConfiguration(configuration['name'], json.dumps(configuration)), exc)

    def _get_location(self, aug, configuration):
        """Return a location instance based on configuration."""

        if not configuration:
            return None
        else:
            factory = self._get_factory(configuration)

            try:
                return factory(self._output, aug, configuration, self._processor_factory)
            except ToolException:
                raise
            except Exception as exc:
                six.raise_from(RookInvalidObjectConfiguration(configuration['name'], json.dumps(configuration)), exc)

    def _get_factory(self, configuration):
        """Return a class factory on configuration."""
        try:
            name = configuration['name']
        except KeyError as exc:
            six.raise_from(RookObjectNameMissing(json.dumps(configuration)), exc)

        try:
            return self.get_object_factory(name)
        except (RookUnknownObject, AttributeError, KeyError) as exc:
            six.raise_from(RookUnsupportedLocation(name), exc)




class Return(object):

    def __init__(self, arguments, factory):
        self._path = factory.get_path(arguments['path'])

    def execute(self, namespace):
        return self._path.read_from(namespace)



from .namespace import Namespace
from .container_namespace import ContainerNamespace
from .python_object_namespace import PythonObjectNamespace
from ...services.monitor.monitor_service import MonitorService

class ProcessStateNamespace(Namespace):
    def __init__(self):
        super(ProcessStateNamespace, self).__init__(self.METHODS)

    def dump(self, args):
        usage = MonitorService.get_last_usage()
        if not usage:
            return ContainerNamespace({})

        return ContainerNamespace(
            {
                "CPU": PythonObjectNamespace(usage.cpu),
                "GlobalCPU": PythonObjectNamespace(usage.global_cpu),
                "VirtualMemorySize": PythonObjectNamespace(usage.virtual_memory),
                "ProcessStartTime": PythonObjectNamespace(usage.start_time),
                "SystemUpTime": PythonObjectNamespace(usage.uptime),
            }
        )

    METHODS = (dump, )



import logging
import time
import six
from rook import config
from rook.user_warnings import UserWarnings
from rook.com_ws.token_bucket import TokenBucket
from rook.protobuf.messages_pb2 import RuleStatusMessage, AugReportMessage, LogMessage
from rook.protobuf.variant_pb2 import Error
from rook.processor.namespace_serializer import NamespaceSerializer
from rook.processor.error import Error as RookError
from rook.processor.namespaces.container_namespace import ContainerNamespace
from rook.logger import logger


class Output(object):
    def __init__(self, agent_id):
        self._id = agent_id
        self._agent_com = None

        self._rule_status_updates_bucket = TokenBucket(config.OutputWsConfiguration.MAX_STATUS_UPDATES,
                                                       config.OutputWsConfiguration.BUCKET_REFRESH_RATE,
                                                       lambda: logger.error("Limit reached, dropping status updates"))

        self._user_message_bucket = TokenBucket(config.OutputWsConfiguration.MAX_AUG_MESSAGES,
                                                config.OutputWsConfiguration.BUCKET_REFRESH_RATE,
                                                lambda: logger.error("Limit reached, dropping aug report messages"))

        self._log_message_bucket = TokenBucket(config.OutputWsConfiguration.MAX_LOG_ITEMS,
                                               config.OutputWsConfiguration.BUCKET_REFRESH_RATE,
                                               lambda: self._internal_send_log_message(3,
                                                                                       time.time(),
                                                                                       __file__,
                                                                                       0,
                                                                                       "Limit reached, dropping log messages",
                                                                                       "Limit reached, dropping log messages"))
        logger.register_output(self)

    def set_agent_id(self, agent_id):
        self._id = agent_id

    def set_agent_com(self, agent_com):
        self._agent_com = agent_com

    def send_rule_status(self, rule_id, active, error):
        if not self._agent_com:
            return

        def send_msg():
            rule_status_message = RuleStatusMessage()
            rule_status_message.agent_id = self._id
            rule_status_message.rule_id = rule_id
            rule_status_message.active = active

            if error:
                rule_status_message.error.CopyFrom(self._convert_error(error.dumps()))

            self._agent_com.add(rule_status_message)
        self._rule_status_updates_bucket.do_if_available(send_msg)

    def send_user_message(self, aug_id, message_id, arguments):
        if not self._agent_com:
            return

        def send_msg():
            if arguments and arguments.size(''):
                if config.DumpConfig.PROTOBUF_VERSION_2:
                    error = self._agent_com.send_user_message(aug_id, message_id, arguments)
                    if error:
                        if not isinstance(error, RookError):
                            error = RookError(error)
                        UserWarnings.send_warning(error)
                else:
                    msg = AugReportMessage()
                    msg.agent_id = self._id
                    msg.aug_id = aug_id
                    msg.report_id = str(message_id)

                    serializer = NamespaceSerializer(config.DumpConfig.STRING_CACHE_USERMESSAGE)
                    serializer.dump(arguments, msg.arguments)

                    for k, v in six.iteritems(serializer.get_string_cache()):
                        msg.strings_cache[k] = v

                    error = self._agent_com.add(msg)
                    if error:
                        UserWarnings.send_warning(RookError(error))
        self._user_message_bucket.do_if_available(send_msg)

    LOG_LEVELS = {
        logging.DEBUG: LogMessage.DEBUG,
        logging.INFO: LogMessage.INFO,
        logging.WARNING: LogMessage.WARNING,
        logging.ERROR: LogMessage.ERROR,
        logging.FATAL: LogMessage.FATAL
    }

    def send_log_message(self, level, time, filename, lineno, text, formatted_message, arguments):
        self._log_message_bucket.do_if_available(
            lambda: self._internal_send_log_message(level, time, filename, lineno, text, formatted_message, arguments)
        )

    def _internal_send_log_message(self, level, time, filename, lineno, text, formatted_message, arguments=None):
        # Until we clean up the initialization of AgentCom & Output (they used
        # to be codependent) we ignore logs from before the rook is actually
        # started
        if self._agent_com is None:
            return

        if arguments is None:
            arguments = {}

        msg = LogMessage()

        msg.timestamp.FromMilliseconds(int(time * 1000))
        msg.agent_id = self._id
        msg.level = level
        msg.filename = filename
        msg.line = lineno
        msg.text = str(text)
        msg.formatted_message = formatted_message
        if arguments:
            NamespaceSerializer().dump(ContainerNamespace(arguments), msg.legacy_arguments)

        self._agent_com.add(msg)

    def send_warning(self, rule_id, error):
        self.send_rule_status(rule_id, "Warning", error)

    def flush_messages(self):
        self._agent_com.flush_all_messages()

    def _convert_error(self, e):
        new_err = Error()
        new_err.message = e.message
        new_err.type = e.type
        new_err.parameters.CopyFrom(e.parameters)
        new_err.exc.CopyFrom(e.exc)
        new_err.traceback.CopyFrom(e.traceback)
        return new_err

    def is_user_messages_queue_full(self):
        return self._user_message_bucket.is_exhausted() or (self._agent_com is not None and self._agent_com.is_queue_full())



import os
from rook.com_ws.agent_com_impel.multithread_agent_com import MultiThreadAgentCom as AgentComMT
from rook.logger import logger
from sys import platform

SINGLE_THREAD_AGENT_COM = object()
MULTI_THREAD_AGENT_COM = object()

thread_connection_type = None

if os.getenv("ROOKOUT_SINGLETHREAD_COMM", "0") == "1":
    if platform in ("linux", "linux2"):
        from rook.com_ws.agent_com_impel.singlethread_agent_com import SingleThreadAgentCom as AgentComSingle

        logger.info("Using single-threaded communication")
        AgentCom = AgentComSingle

        thread_connection_type = SINGLE_THREAD_AGENT_COM
    else:
        logger.info('Not using single-threaded communication because os is not Linux, using multi-threaded communication')
        AgentCom = AgentComMT

        thread_connection_type = MULTI_THREAD_AGENT_COM
else:
    logger.debug("Using multi-threaded communication")
    AgentCom = AgentComMT

    thread_connection_type = MULTI_THREAD_AGENT_COM



import select
import six
from collections import defaultdict

def is_poll_available():
    return hasattr(select, "poll")


if is_poll_available():
    class Waiter(object):
        def __init__(self, rfds, wfds, xfds):
                poller = select.poll()
                self.fds = {x.fileno(): x for x in (rfds + wfds + xfds)}
                flags_per_fd = defaultdict(int)
                for rfd in rfds:
                    flags_per_fd[rfd] = flags_per_fd[rfd] | select.POLLIN | select.POLLPRI
                for wfd in wfds:
                    flags_per_fd[wfd] = flags_per_fd[wfd] | select.POLLOUT
                for xfd in xfds:
                    flags_per_fd[xfd] = flags_per_fd[xfd] | select.POLLHUP | select.POLLERR

                for fd, flags in six.iteritems(flags_per_fd):
                    poller.register(fd, flags)
                self.poller = poller

        def wait(self, timeout):
            rfdsout, wfdsout, xfdsout = [], [], []
            for (fd, bitmask) in self.poller.poll(timeout):
                if bitmask & select.POLLIN or bitmask & select.POLLPRI:
                    rfdsout.append(self.fds[fd])
                if bitmask & select.POLLOUT:
                    wfdsout.append(self.fds[fd])
                if bitmask & select.POLLHUP or bitmask & select.POLLERR:
                    xfdsout.append(self.fds[fd])
            return rfdsout, wfdsout, xfdsout

else:
    class Waiter(object):
        def __init__(self, rfds, wfds, xfds):
            self.rfds, self.wfds, self.xfds = rfds, wfds, xfds

        def wait(self, timeout):
            return select.select(self.rfds, self.wfds, self.xfds, timeout)




import inspect
import os

from .namespace import Namespace
from .python_object_namespace import PythonObjectNamespace


class TracebackNamespace(Namespace):
    def __init__(self, frame, depth):
        super(TracebackNamespace, self).__init__()
        self._frame = frame
        self._depth = depth

    def __getitem__(self, key):
        return self.read_key(key)

    def call_method(self, name, args):
        if name == "size":
            return PythonObjectNamespace(self._depth)
        else:
            return super(TracebackNamespace, self).call_method(name, args)

    def read_key(self, key):
        pos = int(key)

        current_frame = self._frame

        for i in range(pos):
            current_frame = current_frame.f_back()

        return current_frame

    def dump(self, traceback_frames_holder, get_string_index_in_cache=None, increase_pending_size=lambda _size: None):
        """
        Dump traceback frames to given 'traceback_frames_holder'.
        It's the caller responsibility to set the variant type.
        """
        current_frame = self._frame.frame
        lineno = self._frame._lineno

        if get_string_index_in_cache is not None:
            unavailable = get_string_index_in_cache("unavailable")

        for i in range(self._depth):
            frame = traceback_frames_holder.add()

            code = current_frame.f_code
            filename = code.co_filename
            if filename.endswith('.pyc'):
                filename = os.path.splitext(filename)[0] + 'py'
            function = code.co_name

            frame.lineno = lineno if lineno is not None else 0
            if get_string_index_in_cache is None:
                frame.filename = filename if filename is not None else "unavailable"
                frame.name = function if function is not None else "unavailable"
                module = inspect.getmodule(current_frame)
                frame.module = module.__name__ if module is not None else "unavailable"
            else:
                frame.filename_index_in_cache = get_string_index_in_cache(filename) if filename is not None else unavailable
                frame.name_index_in_cache = get_string_index_in_cache(function) if function is not None else unavailable
                frame.module_index_in_cache = unavailable
                increase_pending_size(11)

            current_frame = current_frame.f_back
            if not current_frame:
                break

            lineno = inspect.getlineno(current_frame)



import six
import datetime

from .namespace import Namespace

from rook.exceptions import RookKeyNotFound, RookAttributeNotFound, RookInvalidMethodArguments


class PythonObjectNamespace(Namespace):
    try:
        BINARY_TYPES = (buffer, bytearray)
    except NameError:
        BINARY_TYPES = (bytearray, bytes)

    class ObjectDumpConfig(object):

        STRICT_MAX_DEPTH = 2
        STRICT_MAX_WIDTH = 10
        STRICT_MAX_COLLECTION_DEPTH = 2
        STRICT_MAX_STRING = 128

        DEFAULT_MAX_DEPTH = 4
        DEFAULT_MAX_WIDTH = 10
        DEFAULT_MAX_COLLECTION_DEPTH = 4
        DEFAULT_MAX_STRING = 512

        TOLERANT_MAX_DEPTH = 5
        TOLERANT_MAX_WIDTH = 20
        TOLERANT_MAX_COLLECTION_DEPTH = 5
        TOLERANT_MAX_STRING = 4 * 1024

        UNLIMITED_STRING = 64 * 1024
        UNLIMITED_COLLECTION_WIDTH = 100

        def __init__(self, max_depth=None, max_width=None, max_collection_dump=None, max_string=None):
            self.max_depth = max_depth or self.DEFAULT_MAX_DEPTH
            self.max_width = max_width or self.DEFAULT_MAX_WIDTH
            self.max_collection_dump = max_collection_dump or self.DEFAULT_MAX_COLLECTION_DEPTH
            self.max_string = max_string or self.DEFAULT_MAX_STRING

        def __eq__(self, other):
            return type(self) == type(other) and self.max_depth == other.max_depth and \
                   self.max_width == other.max_width and self.max_collection_dump == other.max_collection_dump and \
                   self.max_string == other.max_string

        @classmethod
        def strict_limits(cls, obj):
            return cls(
                cls.STRICT_MAX_DEPTH,
                cls.STRICT_MAX_WIDTH,
                cls.STRICT_MAX_COLLECTION_DEPTH,
                cls.STRICT_MAX_STRING)

        @classmethod
        def default_limits(cls, obj):
            return cls(
                cls.DEFAULT_MAX_DEPTH,
                cls.DEFAULT_MAX_WIDTH,
                cls.DEFAULT_MAX_COLLECTION_DEPTH,
                cls.DEFAULT_MAX_STRING)

        @classmethod
        def tolerant_limits(cls, obj):
            return cls(
                cls.TOLERANT_MAX_DEPTH,
                cls.TOLERANT_MAX_WIDTH,
                cls.TOLERANT_MAX_COLLECTION_DEPTH,
                cls.TOLERANT_MAX_STRING)

        @classmethod
        def tailor_limits(cls, obj):
            if isinstance(obj, six.string_types):
                return PythonObjectNamespace.ObjectDumpConfig(1, 0, 0, cls.UNLIMITED_STRING)
            if isinstance(obj, (list, dict, set, tuple)) and len(obj) > cls.TOLERANT_MAX_WIDTH:
                return PythonObjectNamespace.ObjectDumpConfig(max_width=cls.UNLIMITED_COLLECTION_WIDTH)
            else:
                return PythonObjectNamespace.ObjectDumpConfig.tolerant_limits(obj)

    def __init__(self, obj, dump_config=None, methods=()):
        super(PythonObjectNamespace, self).__init__(methods + self.METHODS)
        self.obj = obj
        self.dump_config = dump_config or self.ObjectDumpConfig()

    @classmethod
    def get_common_type(cls, obj):
        # NOTE: This list is partially duplicated with the list in NamespaceSerializer._load_variant
        if obj is None:
            return u'null'
        elif isinstance(obj, six.integer_types):
            return u'int'
        elif isinstance(obj, six.string_types):
            return u'string'
        elif isinstance(obj, float):
            return u'float'
        elif isinstance(obj, cls.BINARY_TYPES):
            return u'buffer'
        elif isinstance(obj, datetime.datetime):
            return u'datetime'
        elif isinstance(obj, (list, set, tuple)):
            from .collection_namespace import ListNamespace
            return ListNamespace.get_common_type(obj)
        elif isinstance(obj, (dict)):
            return u'dict'
        elif isinstance(obj, complex):
            return u'complex'
        else:
            raise ValueError("Unknown type- %s", str(type(obj)))

    def read_attribute(self, name):
        try:
            if hasattr(self.obj, '__dict__') and self.obj.__dict__ and name in self.obj.__dict__:
                return PythonObjectNamespace(self.obj.__dict__[name])
            if hasattr(self.obj, '__slots__') and self.obj.__slots__ and name in self.obj.__slots__:
                return PythonObjectNamespace(self.obj.__slots__[name])

            raise RookAttributeNotFound(name)
        except AttributeError:
            raise RookAttributeNotFound(name)

    def read_key(self, key):
        try:
            return PythonObjectNamespace(self.obj[key])
        except (KeyError, IndexError, TypeError):
            # If key is a string and object is a dictionary check if any key within the dictionary stringifies to key
            if isinstance(key, six.string_types) and isinstance(self.obj, dict):
                for iteratorKey in self.obj:
                    if str(iteratorKey) == key:  # The str() is the important part
                        return PythonObjectNamespace(self.obj[iteratorKey])

            raise RookKeyNotFound(key)

    def type(self, args=None):
        return PythonObjectNamespace(self.serialize_type())

    def serialize_type(self):
        return str(type(self.obj))

    def size(self, args=None):
        return PythonObjectNamespace(len(self.obj))

    def depth(self, args):
        try:
            self.dump_config.max_depth = int(args)
        except ValueError:
            raise RookInvalidMethodArguments('depth()', args)

        return self

    def width(self, args):
        try:
            self.dump_config.max_width = int(args)
        except ValueError:
            raise RookInvalidMethodArguments('width()', args)

        return self

    def collection_dump(self, args):
        try:
            self.dump_config.max_collection_dump = int(args)
        except ValueError:
            raise RookInvalidMethodArguments('collection_dump()', args)

        return self

    def string(self, args):
        try:
            self.dump_config.max_string = int(args)
        except ValueError:
            raise RookInvalidMethodArguments('string()', args)

        return self

    def limit(self, args):
        try:
            self.dump_config = PythonObjectNamespace.dump_configs[args.lower()]
        except KeyError:
            raise RookInvalidMethodArguments('limit()', args)

        return self

    def __hash__(self):
        return hash(self.obj)

    def __eq__(self, other):
        return type(self) == other and self.obj == other.obj

    def __nonzero__(self):
        return bool(self.obj)

    if six.PY3:
        __bool__ = __nonzero__

    dump_configs = {
        '': ObjectDumpConfig.default_limits(ObjectDumpConfig),
        'strict': ObjectDumpConfig.strict_limits(ObjectDumpConfig),
        'default': ObjectDumpConfig.default_limits(ObjectDumpConfig),
        'tolerant': ObjectDumpConfig.tolerant_limits(ObjectDumpConfig)}

    METHODS = (type, size, depth, width, collection_dump, string, limit)



from rook.protobuf import protoc

if protoc == "protoc311":
    from rook.protobuf.protoc311.agent_info_pb2 import *
elif protoc == "protoc319":
    from rook.protobuf.protoc319.agent_info_pb2 import *
